/*
 *	Label management
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#define _LABEL_C_
#include "label.h"

char *errmsg;

/* label buffer container */
label_container *label_container_list = 0;
label_container *cl;
label_container *tail;
int namespace_id_serial = 0;
int namespace_id_pass2  = 0;

/* string buffer */
int strbufsiz;
char *strbuf = 0;
int nextstrindex;

/* string buffer */
int lnsbufsiz;
label_across_namespace *lnsbuf = 0;
int nextlnsindex;

int init_labeltable(void) {
	errmsg = "out of memory";

	strbufsiz = DEFSBUFSIZ;
	strbuf = malloc(strbufsiz);
	if (!strbuf) {
	    return(0);
	}
	nextstrindex = 1;	/* 0 should be kept unused */

	cl = new_label_container();
	if (!cl) {
	    free(strbuf);
	    return(0);
	}
	label_container_list = cl;
	tail = cl;

	lnsbufsiz = DEFLNSBUFSIZ;
	lnsbuf = malloc(sizeof(label_across_namespace)*lnsbufsiz);
	if (!lnsbuf) {
	    return(0);
	}
	nextlnsindex = 1;	/* 0 should be kept unused */

	return 1;
}

void tini_labeltable(void) {
	if (label_container_list) dispose_label_containers(label_container_list);
	if (strbuf) free(strbuf);
	if (lnsbuf) free(lnsbuf);
}

unsigned long add_label(label *lp) {	/* 0:error  otherwise:index */
	label *p;
	if (cl->nextlabelindex >= cl->labelbufsiz) {
		cl->labelbufsiz *= 2;
		p = realloc(cl->labelbuf, cl->labelbufsiz * (sizeof(label)));
		if (p == NULL) {
			errmsg = "label buffer overflow";
			return(0);
		}
		cl->labelbuf = p;
	}
	p = &(cl->labelbuf[cl->nextlabelindex]);
	if (lp) *p = *lp;
	return cl->nextlabelindex++;
}

unsigned long add_string(char *str) {	/* 0:error  otherwise:index */
	char *p;
	int l, r;
	l = strlen(str) + 1; /* length + EOS */
	if ((nextstrindex + l) >= strbufsiz) {
		strbufsiz *= 2;
		p = realloc(strbuf, strbufsiz);
		if (p == NULL) {
			errmsg = "string buffer overflow";
			return(0);
		}
		strbuf = p;
	}
	p = &strbuf[nextstrindex];
	strcpy(p, str);
	r = nextstrindex;
	nextstrindex += l;
	return r;
}

int gethash(char *str) {
	char c;
	unsigned long h;
	h = 0;
	while ((c = *str++) != 0) {
		c &= 0x7f;
		c -= 0x10;
		h = (((h >> 25) & 0x7f) | (h << 7)) ^ (long)c;
	}
	return (int)(h % HASHNUM);
}

long getlabel(char *lstr) {
	label *lp;
	unsigned long ni;
	for (ni = cl->htbl[gethash(lstr)]; ni; ni = lp->nextlabel) {
		lp = &(cl->labelbuf[ni]);
		if (!strcmp(&strbuf[lp->labelstr], lstr)) return ni;
	}
	return 0;
}

long entrylabel(char *lstr) {
	label *lp;
	int h;
	unsigned long si, li;
	if (getlabel(lstr)) {
		errmsg = "duplicate label definition";
		return(0);
	}
	h = gethash(lstr);
	if (!(si = add_string(lstr))) return(0);
	if (!(li = add_label(0))) return(0);
	lp = &(cl->labelbuf[li]);
	(void) memset((char *)lp, 0, sizeof(label));
	lp->labelstr = si;
	lp->nextlabel = 0;
	if (cl->htbl[h]) lp->nextlabel = cl->htbl[h];
	cl->htbl[h] = li;
/*dbg fprintf(stderr,"str:%s hash:%d si:%d li:%d\n",lstr,h,si,li);*/
	return li;
}
long entrylabel_across_namespace(char *lstr, long labelid) {
	long li, lir;
	if (lstr == NULL && labelid <= 0) return 0;
	if (lstr == NULL) {
	    li = labelid;
	} else {
	    li = entrylabel(lstr);
	}
	if (!li) return 0;
	if (nextlnsindex >= lnsbufsiz) {
	    label_across_namespace *p;
	    lnsbufsiz *= 2;
	    p = realloc(lnsbuf, lnsbufsiz);
	    if (p == NULL) {
		errmsg = "label-across-namespace buffer overflow";
		return(0);
	    }
	    lnsbuf = p;
	}
	lnsbuf[nextlnsindex].ns = get_namespace_id();
	lnsbuf[nextlnsindex].li = li;
	lir = -nextlnsindex;
	nextlnsindex++;
	return lir;
}

char *getlabelstr(long labelid) {
	label *lp;
	if (labelid < 0) {
	    /* label across namespace */
	    lp = getlabelptr(labelid);
	    if (lp == NULL) return NULL;
	    return &strbuf[lp->labelstr];	/* beware lifetime of pointer */
	}
	lp = &(cl->labelbuf[labelid]);
	return &strbuf[lp->labelstr];		/* beware lifetime of pointer */
}
label *getlabelptr(long labelid) {
	label *lp;
	if (labelid < 0) {
	    /* label across namespace */
	    int i, nssav;
	    i = -labelid;
	    if (labelid >= lnsbufsiz) return NULL;
	    nssav = get_namespace_id();
	    if (!switch_namespace(lnsbuf[i].ns)) return NULL;
	    lp = &(cl->labelbuf[lnsbuf[i].li]);	/* beware lifetime of pointer */
	    switch_namespace(nssav);
	    return lp;
	}
	if (labelid < 1 || labelid >= cl->nextlabelindex) return NULL;
	lp = &(cl->labelbuf[labelid]);
	return lp;				/* beware lifetime of pointer */
}

char *get_string(long strid) {
	return &strbuf[strid];
}

label *getlabelptr_outside(char *lstr, char dstt) {
	/* search upper namespaces */
	label *lp;
	int nssav;
	long li;
	nssav = get_namespace_id();
	while (leave_namespace()) {
	    li = getlabel(lstr);
	    if (!li) continue;
	    lp = getlabelptr(li);
	    if (lp->defstate == dstt) {
		switch_namespace(nssav);
	        return lp;
	    }
	}
	switch_namespace(nssav);
	return 0;
}

label_container *new_label_container(void) {
	label_container *lcp;
	int i;

	errmsg = "out of memory";
	lcp = malloc(sizeof(label_container));
	if (!lcp) return 0;
	lcp->labelbufsiz = DEFLBUFSIZ;
	lcp->labelbuf = malloc(DEFLBUFSIZ * (sizeof(label)));
	if (!lcp->labelbuf) {
	    free(lcp);
	    return(0);
	}
	for (i=0; i<HASHNUM; i++) lcp->htbl[i] = 0L;
	lcp->nextlabelindex = 1;	/* 0 should be kept unused */
	lcp->parent = 0;
	lcp->next   = 0;
	lcp->name[0] = 0;
	lcp->id = namespace_id_serial++;

	return lcp;
}

void dispose_label_containers(label_container *lcp) {
	label_container *lcp0;
	while (lcp) {
	    lcp0 = lcp->next;
	    free(lcp->labelbuf);
	    free(lcp);
	    lcp = lcp0;
	}
}

int get_namespace_id(void) {
	return cl->id;
}
int switch_namespace(int id) {
	label_container *lcp;
	for (lcp = label_container_list; lcp; lcp = lcp->next) {
	    if (lcp->id == id) {
		cl = lcp;
		return 1;
	    }
	}
	return 0;
}

int dig_namespace(char *namespace_name) {
	label_container *lcp;
	lcp = new_label_container();
	if (!lcp) return 0;
	if (namespace_name) strcpy(lcp->name, namespace_name);
	lcp->parent = cl;
	tail->next = lcp;
	tail = lcp;
	cl = lcp;
	return 1;
}
void dig_namespace_pass2(void) {
	switch_namespace(++namespace_id_pass2);
}

int leave_namespace(void) {
	if (cl->parent) {
	    cl = cl->parent;
	    return 1;
	}
	return 0;
}
void topmost_namespace(void) {
	while (leave_namespace());
}
int next_namespace(void) {
	if (cl->next) {
	    cl = cl->next;
	    return 1;
	}
	return 0;
}
char *getlabelstr_with_namespace(long labelid) {
	static char buf[256];
	char *p;
	label *lp;
	label_container *lpc;
	int t = 256;
	int l;
	lp = &(cl->labelbuf[labelid]);
	p = &strbuf[lp->labelstr];
	l = strlen(p) + 1;
        t -= l;
	strcpy(&buf[t], p);
	for (lpc = cl; lpc->parent; lpc = lpc->parent) {
	    l = strlen(lpc->name);
	    if (t<=l) {
		break;
	    }
	    buf[--t] = '.';
	    t -= l;
	    strncpy(&buf[t], lpc->name, l);
	}
	return &buf[t];
}
