/*
 *	Z80 assembler parsing routines
 */

#define _PARSE_C_

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "parse.h"
#include "label.h"
#include "token.h"
#include "linebuf.h"

extern int parse_text(parseinfo*);

static char* rp, * jp;
static int loccnt;	// current offset in binary buffer
static int locabs;	// abusolute address of loccnt
static int locdol;	// address referred by '$'
static int sav_locabs;
static int sav_loccnt;
static int loccnt_start;
static int locabs_start;

static int exec_addr;
static int load_addr;

#define MAXBINBUF 65536
static char binbuf[MAXBINBUF * 2];

static int locallabel[11];

static int macro_nest_lev;
#define MAXMACRONEST 4

static int rept_num;

#define DB_ASCII_GENERIC 0
#define DB_ASCII_MZ_JP   1
#define DB_ASCII_MZ_EU   2
#define DB_DISP_JP       3
#define DB_DISP_EU       4
#define DB_USER_DEFINED  5
static int db_encode;

static int arr_num;
static int arr_index;
static int arr_count;
static int arr_bytes[64];

static long using_in_page;
static int loccnt_in_page;
static int using_in_page_type;
static int using_in_page_upto256;
static int using_in_page_used;

static int pass2;

static char* errmsg;
static char emstr[256];

static char mapchr[256];

char* mjrerr[] = {
	EJ("Syntax error","\G["),
#define EM_SYNTAX_ERROR		0
	EJ("Illegal operand","ȃIyh"),
#define EM_ILLEGAL_OPERAND	1
	EJ("Unresolved expression","̒lm"),
#define EM_UNRESOLVED_EXPR	2
	EJ("Out of range on relative jump","͈͊Oւ̑΃Wv"),
#define EM_JR_OVERFLOW		3
	EJ("Duplicate label definition","x̓d`"),
#define EM_DUP_LABEL		4
	EJ("Missing ')'","')'܂"),
#define EM_MISSING_RBRC		5
	EJ("Bit count should be in 0-7","rbgw肪͈͊O"),
#define EM_BITCNT_OVERFLOW	6
	EJ("Missing opening '\"'","Jn'\"'܂"),
#define EM_MISSING_OPEN_DC	7
	EJ("Missing closing '\"'","I'\"'܂"),
#define EM_MISSING_CLOSE_DC	8
	EJ("Closing quote is missing","I'''܂"),
#define EM_MISSING_CLOSE_QUOTE	9
	EJ("Something is after %s on the same line","%š͍s܂ŉLqł܂"),
#define EM_SOMETHING_TAILED	10
	"Too much IF-ELSE-ENDIF nesting",
#define EM_IF_TOO_NESTED	11
	0
};

void em_dup_label(char* lstr, char dstt) {
	sprintf(emstr, EJ("%s is already defined as a %s", "%s͂ł%sƂĒ`Ă܂"), lstr,
		dstt == LABEL_DEFINED ? EJ("label", "x") :
		dstt == LABEL_MACRO ? EJ("macro", "}N") : EJ("something", "???"));
	errmsg = emstr;
}

/* ASCII to MZ-700 Display code table */
static char dispc[96] = {
	/* SP    !     "     #     $     %     &     '     (     )     *     +     ,     -     .     / */
  0x00, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69, 0x6b, 0x6a, 0x2f, 0x2a, 0x2e, 0x2d,
  /* 0     1     2     3     4     5     6     7     8     9     :     ;     <     =     >     ? */
0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27, 0x28, 0x29, 0x4f, 0x2c, 0x51, 0x2b, 0x57, 0x49,
/* @     A     B     C     D     E     F     G     H     I     J     K     L     M     N     O */
0x55, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f,
/* P     Q     R     S     T     U     V     W     X     Y     Z     [     \     ]     ^     _ */
0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17, 0x18, 0x19, 0x1a, 0x52, 0x59, 0x54, 0x50, 0x3c,
/* `     a     b     c     d     e     f     g     h     i     j     k     l     m     n     o */
0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f,
/* p     q     r     s     t     u     v     w     x     y     z     {     |     }     ~       */
0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17, 0x18, 0x19, 0x1a, 0x00, 0x79, 0x00, 0xe3, 0
};
static char dispc_eu[96] = {
	/* SP    !     "     #     $     %     &     '     (     )     *     +     ,     -     .     / */
  0x00, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69, 0x6b, 0x6a, 0x2f, 0x2a, 0x2e, 0x2d,
  /* 0     1     2     3     4     5     6     7     8     9     :     ;     <     =     >     ? */
0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27, 0x28, 0x29, 0x4f, 0x2c, 0x51, 0x2b, 0x57, 0x49,
/* @     A     B     C     D     E     F     G     H     I     J     K     L     M     N     O */
0x55, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f,
/* P     Q     R     S     T     U     V     W     X     Y     Z     [     \     ]     ^     _ */
0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17, 0x18, 0x19, 0x1a, 0x52, 0x59, 0x54, 0xbe, 0xbf,
/* `     a     b     c     d     e     f     g     h     i     j     k     l     m     n     o */
0xa4, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89, 0x8a, 0x8b, 0x8c, 0x8d, 0x8e, 0x8f,
/* p     q     r     s     t     u     v     w     x     y     z     {     |     }     ~       */
0x90, 0x91, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98, 0x99, 0x9a, 0xbc, 0x80, 0x40, 0xa5, 0
};
static char dispck[64] = {
	0x00, 0xbd, 0x9d, 0xb1, 0xb5, 0xb9, 0xb4, 0x9e, 0xb2, 0xb6, 0xba, 0xbe, 0x9f, 0xb3, 0xb7, 0xbb,
	0xbf, 0xa3, 0x85, 0xa4, 0xa5, 0xa6, 0x94, 0x87, 0x88, 0x9c, 0x82, 0x98, 0x84, 0x92, 0x90, 0x83,
	0x91, 0x81, 0x9a, 0x97, 0x93, 0x95, 0x89, 0xa1, 0xaf, 0x8b, 0x86, 0x96, 0xa2, 0xab, 0xaa, 0x8a,
	0x8e, 0xb0, 0xad, 0x8d, 0xa7, 0xa8, 0xa9, 0x8f, 0x8c, 0xae, 0xac, 0x9b, 0xa0, 0x99, 0xbc, 0xb8
};
static char euasc[32] = {
		0x93, 0xA1, 0x9A, 0x9F, 0x9C, 0x92, 0xAA, 0x97, 0x98, 0xA6, 0xAF, 0xA9, 0xB8, 0xB3, 0xB0, 0xB7,
		0x9E, 0xA0, 0x9D, 0xA4, 0x96, 0xA5, 0xAB, 0xA3, 0x9B, 0xBD, 0xA2, 0xBE, 0xC0, 0x80, 0x94, 0x7F
};

//struct ttbl {
//    int	moji;
//    int	dispc;
//};
//struct ttbl zen2dispc[] = {
//    { '', 0xd0 },  { '', 0xd1 },  { '', 0xd2 },  { '', 0xd3 },
//    { '', 0xd4 },  { '', 0xd5 },  { 'y', 0xd6 },  { '', 0xd7 },
//    { 'N', 0xd8 },  { '', 0xd9 },  { '', 0xda },  { 'b', 0xdb },
//    { '', 0xdd },  { '', 0xde },  { '~', 0xdc },
//    { '', 0x50 },  { '', 0x45 },  { '', 0x40 },  { '', 0x80 },
//    { 'B', 0xbd },  { 'u', 0x9d },  { 'v', 0xb1 },  { 'A', 0xb5 },  { 'D', 0xb9 },
//    { '', 0xb4 },  
//    { '@', 0x9e },  { 'B', 0xb2 },  { 'D', 0xb6 },  { 'F', 0xba },  { 'H', 0xbe },
//    { '', 0x9f },  { '', 0xb3 },  { '', 0xb7 },  { 'b', 0xbb },  { '[', 0xbf },
//    { 'A', 0xa3 },  { 'C', 0x85 },  { 'E', 0xa4 },  { 'G', 0xa5 },  { 'I', 0xa6 },
//    { 'J', 0x94 },  { 'L', 0x87 },  { 'N', 0x88 },  { 'P', 0x9c },  { 'R', 0x82 },
//    { 'T', 0x98 },  { 'V', 0x84 },  { 'X', 0x92 },  { 'Z', 0x90 },  { '\', 0x83 },
//    { '^', 0x91 },  { '`', 0x81 },  { 'c', 0x9a },  { 'e', 0x97 },  { 'g', 0x93 },
//    { 'i', 0x95 },  { 'j', 0x89 },  { 'k', 0xa1 },  { 'l', 0xaf },  { 'm', 0x8b },
//    { 'n', 0x86 },  { 'q', 0x96 },  { 't', 0xa2 },  { 'w', 0xab },  { 'z', 0xaa },
//    { '}', 0x8a },  { '~', 0x8e },  { '', 0xb0 },  { '', 0xad },  { '', 0x8d },
//    { '', 0xa7 },  { '', 0xa8 },  { '', 0xa9 },
//    { '', 0x8f },  { '', 0x8c },  { '', 0xae },  { '', 0xac },  { '', 0x9b },
//    { '', 0xa0 },  { '', 0x99 },  { 'J', 0xbc },  { 'K', 0xb8 },
//    { 'K', 0x194 },  { 'M', 0x187 },  { 'O', 0x188 },  { 'Q', 0x19c },  { 'S', 0x182 },
//    { 'U', 0x198 },  { 'W', 0x184 },  { 'Y', 0x192 },  { '[', 0x190 },  { ']', 0x183 },
//    { '_', 0x191 },  { 'a', 0x181 },  { 'd', 0x19a },  { 'f', 0x197 },  { 'h', 0x193 },
//    { 'o', 0x186 },  { 'r', 0x196 },  { 'u', 0x1a2 },  { 'x', 0x1ab },  { '{', 0x1aa },
//    { 'p', 0x286 },  { 's', 0x296 },  { 'v', 0x2a2 },  { 'y', 0x2ab },  { '|', 0x2aa },
//    { '', 0xb4 },  
//    { '', 0x9e },  { '', 0xb2 },  { '', 0xb6 },  { '', 0xba },  { '', 0xbe },
//    { '', 0x9f },  { '', 0xb3 },  { '', 0xb7 },  { '', 0xbb },
//    { '', 0xa3 },  { '', 0x85 },  { '', 0xa4 },  { '', 0xa5 },  { '', 0xa6 },
//    { '', 0x94 },  { '', 0x87 },  { '', 0x88 },  { '', 0x9c },  { '', 0x82 },
//    { '', 0x98 },  { '', 0x84 },  { '', 0x92 },  { '', 0x90 },  { '', 0x83 },
//    { '', 0x91 },  { '', 0x81 },  { '', 0x9a },  { '', 0x97 },  { '', 0x93 },
//    { '', 0x95 },  { '', 0x89 },  { '', 0xa1 },  { '', 0xaf },  { '', 0x8b },
//    { '', 0x86 },  { '', 0x96 },  { '', 0xa2 },  { '', 0xab },  { '', 0xaa },
//    { '', 0x8a },  { '', 0x8e },  { '', 0xb0 },  { '', 0xad },  { '', 0x8d },
//    { '', 0xa7 },  { '', 0xa8 },  { '', 0xa9 },
//    { '', 0x8f },  { '', 0x8c },  { '', 0xae },  { '', 0xac },  { '', 0x9b },
//    { '', 0xa0 },  { '', 0x99 },
//    { '', 0x194 },  { '', 0x187 },  { '', 0x188 },  { '', 0x19c },  { '', 0x182 },
//    { '', 0x198 },  { '', 0x184 },  { '', 0x192 },  { '', 0x190 },  { '', 0x183 },
//    { '', 0x191 },  { '', 0x181 },  { '', 0x19a },  { '', 0x197 },  { '', 0x193 },
//    { '', 0x186 },  { '', 0x196 },  { '', 0x1a2 },  { '', 0x1ab },  { '', 0x1aa },
//    { '', 0x286 },  { '', 0x296 },  { '', 0x2a2 },  { '', 0x2ab },  { '', 0x2aa },
//    { '', 0x1e },  { '', 0x1c },  { '', 0x5d },  { '', 0x5c },  { '', 0x1f },
//    { '', 0x5f },  { '', 0x5e },  { '', 0x1d },  { '', 0x78 },  { '', 0x79 },
//    { '', 0x1b },  { '~', 0x6d },  { '_', 0x77 },  { '^', 0x76 },
//    { '', 0x47 },  { '', 0x48 },  { '', 0x44 },  { '', 0x43 },
//    { '', 0x60 },
//    { 0, 0 }
//};

static parseinfo* pinfo;

typedef struct unres_expr_t {
	struct unres_expr_t* next;
	int	patchloc;	// offset to patch
	int	patchtype;	// Label:0 Byte:1 Word:2 RelJump:3
#define PATCH_LABEL	0
#define PATCH_BYTE	1
#define PATCH_WORD	2
#define PATCH_RELJMP	3
#define PATCH_BITCNT	4
	int	linenum;	// line number
	int	depth;
	int	namespaceid;	// id of the namespace of this scope
	long	filename;	// id of filename string
	long	vstack[1];
} unres_expr;

/*---------------------------------------*/

int skipspace(void) { /* 1:skipped chars are all spaces, 0:includes comment */
	char c;
	for (;;) {
		c = *rp;
		if (isspace(c)) {
			rp++;
			continue;
		}
		if (c == ';') {
			/* comment */
			do { rp++; } while (*rp);
			return 0;
		}
		break;
	}
	return 1;
}
void skip_until_eol(void) {
	while (*rp) rp++;
}

char peek1chr(void) {
	return *rp;
}
char peek1chrtu(void) {
	return (char)toupper((int)*rp);
}
char read1chr(void) {
	char c;
	c = peek1chr();
	if (c) rp++;
	return c;
}
char read1chrtu(void) {
	char c;
	c = peek1chrtu();
	if (c) rp++;
	return c;
}
int match1chrtu(char c0) {
	char c;
	c = peek1chrtu();
	if (c == c0) {
		rp++;
		return 1;
	}
	return 0;
}

int isIDchar(char c) {
	return (isalnum(c) || c == '_' || c == '#' || c == '?');
}
int is1stIDchar(char c) {
	return (isalpha(c) || c == '_' || c == '#' || c == '?');
}

/*
	Tables of identifier
 */
typedef struct {
	long	lname;
	int	id;
} idtbl;

idtbl id1tbl[] = {
	{ 'A', REG_A }, { 'B', REG_B },{ 'C', REG_C },{ 'D', REG_D },
	{ 'E', REG_E }, { 'H', REG_H },{ 'L', REG_L },{ 'Z', CND_Z },
	{ 'R', REG_R }, { 'I', REG_I },{ 'M', CND_M },{ 'P', CND_P },
	{ 'V', CND_V },
	{ 0, 0 }
};

idtbl id2tbl[] = {
	{ 'AF', REG_AF }, { 'BC', REG_BC }, { 'DE', REG_DE }, { 'HL', REG_HL },
	{ 'SP', REG_SP }, { 'IX', REG_IX }, { 'IY', REG_IY },
	{ 'LD', INST_LD }, { 'JR', INST_JR }, { 'CP', INST_CP }, { 'OR', INST_OR },
	{ 'JP', INST_JP }, { 'EX', INST_EX }, { 'IN', INST_IN }, { 'RL', INST_RL },
	{ 'RR', INST_RR }, { 'DI', INST_DI }, { 'EI', INST_EI }, { 'IM', INST_IM },
	{ 'NC', CND_NC }, { 'NZ', CND_NZ }, { 'PO', CND_PO }, { 'PE', CND_PE },
	{ 'NV', CND_NV },
	{ 'DB', CMD_DB },	/* pseudo */
	{ 'DW', CMD_DW },	/* pseudo */
	{ 'DS', CMD_DS },	/* pseudo */
	{ 'IF', CMD_IF },	/* pseudo */
	{ 0, 0 }
};

idtbl id3tbl[] = {
	{ 'IXH', REG_IXH }, { 'IXL', REG_IXL }, { 'IYH', REG_IYH }, { 'IYL', REG_IYL },
	{ 'INC', INST_INC }, { 'DEC', INST_DEC }, { 'ADD', INST_ADD }, { 'ADC', INST_ADC },
	{ 'SUB', INST_SUB }, { 'SBC', INST_SBC }, { 'AND', INST_AND }, { 'XOR', INST_XOR },
	{ 'RET', INST_RET }, { 'POP', INST_POP }, { 'NOP', INST_NOP }, { 'RLA', INST_RLA },
	{ 'RRA', INST_RRA }, { 'CPL', INST_CPL }, { 'EXX', INST_EXX }, { 'OUT', INST_OUT },
	{ 'NEG', INST_NEG }, { 'LDI', INST_LDI }, { 'LDD', INST_LDD }, { 'SCF', INST_SCF },
	{ 'CCF', INST_CCF }, { 'RLC', INST_RLC }, { 'RRC', INST_RRC }, { 'SLA', INST_SLA },
	{ 'SRA', INST_SRA }, { 'SLL', INST_SLL }, { 'SRL', INST_SRL }, { 'BIT', INST_BIT },
	{ 'RES', INST_RES }, { 'SET', INST_SET }, { 'RST', INST_RST }, { 'CPI', INST_CPI },
	{ 'CPD', INST_CPD }, { 'INI', INST_INI }, { 'IND', INST_IND }, { 'DAA', INST_DAA },
	{ 'RRD', INST_RRD }, { 'RLD', INST_RLD },
	{ 'EQU', CMD_EQU },	/* pseudo */
	{ 'ORG', CMD_ORG },	/* pseudo */
	{ 'END', CMD_END },	/* pseudo */
	{ 0, 0 }
};

idtbl id4tbl[] = {
	{ 'DJNZ', INST_DJNZ }, { 'RLCA', INST_RLCA }, { 'RRCA', INST_RRCA }, { 'HALT', INST_HALT },
	{ 'CALL', INST_CALL }, { 'PUSH', INST_PUSH }, { 'RETN', INST_RETN }, { 'RETI', INST_RETI },
	{ 'LDIR', INST_LDIR }, { 'LDDR', INST_LDDR }, { 'CPIR', INST_CPIR }, { 'CPDR', INST_CPDR },
	{ 'INIR', INST_INIR }, { 'INDR', INST_INDR }, { 'OUTI', INST_OUTI }, { 'OUTD', INST_OUTD },
	{ 'OTIR', INST_OTIR }, { 'OTDR', INST_OTDR }, { 'TSTI', INST_TSTI },
	{ 'ENDM', CMD_ENDM },	/* pseudo */
	{ 'ELSE', CMD_ELSE },	/* pseudo */
	{ 'REPT', CMD_REPT },	/* pseudo */
	{ 'DEFB', CMD_DB },	/* pseudo */
	{ 'DEFW', CMD_DW },	/* pseudo */
	{ 'DEFS', CMD_DS },	/* pseudo */
	{ 'PROC', CMD_PROC },	/* pseudo */
	{ 'ENDP', CMD_ENDP },	/* pseudo */
	{ 'ELIF', CMD_ELIF },	/* pseudo */
	{ 0, 0 }
};

long fourc2long(char* p) {
	long l = 0;
	int i;
	for (i = 0; i < 4; i++) {
		if (p[i]) {
#if (('0123' & 0x000000FF) == '0')
			l |= (toupper(p[i]) << (i * 8));
#else
			l <<= 8;
			l |= toupper(p[i]);
#endif
		}
		else break;
	}
	return l;
}

int chk_rsvword(token* t) {
	if (t->type != IDENTIFIER) return 0;
	if (t->tstrlen > 6) {
		if (!strcmpi(t->tstr, "INCLUDE")) t->type = CMD_INCLUDE;
		else if (!strcmpi(t->tstr, "DEFINED")) t->type = OPR_DEFINED;
		else if (!strcmpi(t->tstr, "BINCLUDE")) t->type = CMD_BINCLUDE;
		else if (!strcmpi(t->tstr, "ARRANGE")) t->type = CMD_ARRANGE;
		else if (!strcmpi(t->tstr, "IN_PAGE")) t->type = CMD_IN_PAGE;
		else if (!strcmpi(t->tstr, "PAGE_ALIGNED")) t->type = CMD_PAGE_ALIGNED;
		else return 0;
		return 1;
	}
	if (t->tstrlen > 4) {
		if (!strcmpi(t->tstr, "MACRO")) t->type = CMD_MACRO;
		else if (!strcmpi(t->tstr, "ALIGN")) t->type = CMD_ALIGN;
		else if (!strcmpi(t->tstr, "ENDIF")) t->type = CMD_ENDIF;
		else if (!strcmpi(t->tstr, "DISPC")) t->type = CMD_DISPC;
		else if (!strcmpi(t->tstr, "ERROR")) t->type = CMD_ERROR;
		else if (!strcmpi(t->tstr, "ELSIF")) t->type = CMD_ELIF;
		else if (!strcmpi(t->tstr, "OFFSET")) t->type = CMD_OFFSET;
		else if (!strcmpi(t->tstr, "ELSEIF")) t->type = CMD_ELIF;
		else if (!strcmpi(t->tstr, "CHRMAP")) t->type = CMD_CHRMAP;
		else if (!strcmpi(t->tstr, "MAPCHR")) t->type = CMD_MAPCHR;
		else return 0;
		/*dbg	    printf("Reserved Word: %d(%s)\n", t->type, t->tstr);*/
		return 1;
	}
	else {
		long l;
		idtbl* idt;
		static idtbl* aidt[4] = { id1tbl, id2tbl, id3tbl, id4tbl };
		idt = aidt[t->tstrlen - 1];
		l = fourc2long(t->tstr);
		for (; idt->id; idt++) {
			if (l == idt->lname) {
				t->type = idt->id;
				/*dbg		    printf("Reserved Word: %d(%s)\n", t->type, t->tstr);*/
				return 1;
			}
		}
	}
	/*dbg	printf("Identifier:%s\n",t->tstr);*/
	return 0;
}

int long2cstr(char* p, long l) {	// Encode 30bit numeric value to 5-digit string (without EOS)
	int i, j;
	if (l & 0xC0000000) return 0;
	for (i = 0; i < 5; i++) {
		j = (int)((l & 0x3F000000) >> 24);
		if (j < 12) {
			*p++ = (char)(j + '0');			// 0-9, :, ;
		}
		else if (j < (12 + 26)) {
			*p++ = (char)(j - 12 + 'A');		// A-Z
		}
		else /*if (j < (12+26+26))*/ {
			*p++ = (char)(j - (12 + 26) + 'a');	// a-z
		}
		l <<= 6;
	}
	return 1;
}

int gethex(void) {
	int vh = 0;
	char c;
	while (isxdigit(c = peek1chrtu()) || c == '_') {
		read1chr();
		if (c == '_') continue;
		vh <<= 4;
		if (isdigit(c)) vh += (int)(c - '0');
		else	    vh += (int)(c - 'A' + 10);
	}
	return vh;
}

/*
 *	Get Token
 */
int get_token(token* t) {
	char c;

	t->from = rp;
	if (skipspace()) t->from = rp;

	/* if-else-endif handling */
	if (pinfo->ifstt == IFSTT_SKIP_UNTIL_ELSE ||
		pinfo->ifstt == IFSTT_SKIP_UNTIL_ENDIF ||
		pinfo->ifstt == IFSTT_SKIP_ALL) {
		while ((c = read1chrtu()) != 0) {
			if (c == 'E') {
				if (match1chrtu('L')) {
					if (match1chrtu('S') &&
						match1chrtu('E') &&
						!isIDchar(peek1chrtu())) {
						/* 'ELSE' is found */
						t->type = CMD_ELSE;
						return 1;
					}
					else if (match1chrtu('I') &&
						match1chrtu('F') &&
						!isIDchar(peek1chrtu())) {
						/* 'ELIF' is found */
						t->type = CMD_ELIF;
						return 1;
					}
				}
				else if (match1chrtu('N')) {
					if (match1chrtu('D') &&
						match1chrtu('I') &&
						match1chrtu('F') &&
						!isIDchar(peek1chrtu())) {
						/* 'ENDIF' is found */
						t->type = CMD_ENDIF;
						return 1;
					}
				}
				/* skip identifier chars if not match */
				while (isIDchar(peek1chrtu())) read1chr();
			}
			else if (c == 'I') {
				if (read1chrtu() == 'F' &&
					!isIDchar(peek1chrtu())) {
					/* 'IF' is found while skipping... */
					if ((pinfo->ifsp + 1) == MAX_IFSTK) {
						errmsg = mjrerr[EM_IF_TOO_NESTED];
						return 0;
					}
					pinfo->ifstk[(pinfo->ifsp)++] = pinfo->ifstt;
					pinfo->ifstt = IFSTT_SKIP_ALL;
				}
				/* skip identifier chars if not match */
				while (isIDchar(peek1chrtu())) read1chr();
			}
			else if (c == ';') skipspace();
			else if (c == 0x27) /* skip single-quoted string */
				while ((c = peek1chr()) != 0 && c != 0x27) read1chr();
			else if (c == 0x22) /* skip double-quoted string */
				while ((c = peek1chr()) != 0 && c != 0x22) read1chr();
			else if (is1stIDchar(c)) /* skip identifier */
				while ((c = peek1chr()) != 0 && isIDchar(c)) read1chr();
			else /* skip symbols */
				while ((c = peek1chr()) != 0 && !isIDchar(c) &&
					c != ';' && c != 0x22 && c != 0x27) read1chr();
		}
		//	  t->type = 0;/*EOS*/
		t->type = DELTEXT;
		return 1;
	}

	c = peek1chrtu();
	if (is1stIDchar(c)) {
		/* label, instruction, operand */
		int i;
		for (i = 0; i < MAXTOKENLEN - 1; i++) {
			c = peek1chr();
			if (!isIDchar(c)) break;
			t->tstr[i] = c;
			read1chr();
		}
		t->tstr[i] = 0;
		t->tstrlen = i;
		t->type = IDENTIFIER;
		t->value = 0;
		chk_rsvword(t);
		if (t->type == REG_AF && peek1chr() == 0x27) {
			t->type = REG_AF_; /* AF' */
			read1chr();
		}
		return 1;
	}
	else if (c >= '0' && c <= '9') {
		/* immidiate value */
		int vb, vd, vh;
		int fb, fd;
		char c2;
		vb = vd = vh = c - '0';
		fb = (vb <= 1);
		fd = 1;
		read1chr();
		if (c == '0' && peek1chrtu() == 'X') {
			/* hex value (C style) */
			read1chr();
			if (isxdigit(peek1chrtu())) {
				t->type = IMMIDIATE;
				t->value = gethex();
				return 1;
			}
		}
		for (;; read1chr()) {
			c2 = peek1chrtu();
			/* end of hex immidiate? */
			if (c2 == 'H') {
				/* hex value */
				t->type = IMMIDIATE;
				t->value = vh;
				read1chr();
				return 1;
			}
			/* hex digit? */
			if (isxdigit(c2)) {
				vh <<= 4;
				if (isdigit(c2)) vh += (int)(c2 - '0');
				else		 vh += (int)(c2 - 'A' + 10);
				/* if hex digit after '0101B', it is hex value, not binary */
				if (fb == 2) fb = 0;
				/* if 'B' after binary digits, record it */
				if (c2 == 'B' && fb == 1) {
					fb = 2;
					fd = 0;
					continue;
				}
				/* decimal digit? */
				if (fd && (fd = isdigit(c2)) != 0) {
					vd *= 10;
					vd += (int)(c2 - '0');
				}
				/* binary digit? */
				if (fb && (fb = (c2 == '0' || c2 == '1')) != 0) {
					vb <<= 1;
					vb |= (int)(c2 - '0');
				};
				continue;
			}
			/* separator? */
			if (c2 == '_') continue;
			/* not [0-9A-FH_] */
			if (fb == 2) {
				/* binary value */
				t->type = IMMIDIATE;
				t->value = vb;
				return 1;
			}
			if (fd) {
				/* binary value */
				t->type = IMMIDIATE;
				t->value = vd;
				return 1;
			}
			errmsg = EJ("Illegal numeric character", "ȐlLN^");
			return 0;
		}
	}
	else if (c == '$') {
		/* address of this point */
		read1chr();
		if (!isxdigit(c = peek1chrtu())) {
			t->type = T_LOCCNT;
		}
		else {
			/* hex digit */
			t->value = gethex();
			t->type = IMMIDIATE;
		}
		return 1;
	}
	else if (c == ',') {
		/* comma */
		read1chr();
		t->type = T_COMMA;
		return 1;
	}
	else if (c == '(') {
		/* operand, expression */
		char c1, c2, c3;
		char* oldrp;
		read1chr();
		skipspace();
		oldrp = rp;
		c1 = read1chrtu();
		if (c1 == 'C') {
			skipspace();
			if (read1chr() == ')') {
				t->type = REG_C_BRACED;
				return 1;
			}
			goto brace_xit;
		}
		c2 = read1chrtu();
		skipspace();
		c3 = peek1chr();
		if (c1 == 'H' && c2 == 'L' && c3 == ')') {
			t->type = REG_HL_BRACED;
			read1chr();
			return 1;
		}
		else if (c1 == 'D' && c2 == 'E' && c3 == ')') {
			t->type = REG_DE_BRACED;
			read1chr();
			return 1;
		}
		else if (c1 == 'B' && c2 == 'C' && c3 == ')') {
			t->type = REG_BC_BRACED;
			read1chr();
			return 1;
		}
		else if (c1 == 'S' && c2 == 'P' && c3 == ')') {
			t->type = REG_SP_BRACED;
			read1chr();
			return 1;
		}
		else if (c1 == 'I' && c2 == 'X' && c3 == ')') {
			t->type = REG_IX_BRACED;
			read1chr();
			return 1;
		}
		else if (c1 == 'I' && c2 == 'Y' && c3 == ')') {
			t->type = REG_IY_BRACED;
			read1chr();
			return 1;
		}
		else if (c1 == 'I' && c2 == 'X' && (c3 == '+' || c3 == '-')) {
			t->type = REG_IX_d_BRACED;
			return 1;
		}
		else if (c1 == 'I' && c2 == 'Y' && (c3 == '+' || c3 == '-')) {
			t->type = REG_IY_d_BRACED;
			return 1;
		}
	brace_xit:	rp = oldrp;
		t->type = T_BRCOPEN;
		return 1;

	}
	else if (c == ')') {
		/* operand, expression */
		read1chr();
		t->type = T_BRCCLOSE;
		return 1;
	}
	else if (c == '@') {
		/* local label */
		char c1;
		int i;
		read1chr();
		c = peek1chrtu();
		if (c == '@' || c == 'B' || c == 'F') {
			/* '@@', '@b', '@f' */
			if (c == '@') read1chr();
			c1 = c;
			c = '@';
			i = 10;
			goto proc_locallabel;
		}
		else if (c >= '0' && c <= '9') {
			read1chr();
			i = (int)(c - '0');
			c1 = peek1chrtu();
		proc_locallabel:
			if (c1 == 'B') {
				/* '@nB' */
				read1chr();
				sprintf(t->tstr, "@%c00000", c);
				long2cstr(&(t->tstr[2]), locallabel[i] - 1);
				t->type = IDENTIFIER;
				t->tstrlen = 7;
				t->value = IDENTIFIER_REF_ONLY;
				return 1;
			}
			else if (c1 == 'F') {
				/* '@nF' */
				read1chr();
				sprintf(t->tstr, "@%c00000", c);
				long2cstr(&(t->tstr[2]), locallabel[i]);
				t->type = IDENTIFIER;
				t->tstrlen = 7;
				t->value = IDENTIFIER_REF_ONLY;
				return 1;
			}
			else {
				/* '@n' */
				sprintf(t->tstr, "@%c00000", c);
				long2cstr(&(t->tstr[2]), locallabel[i]);
				t->type = IDENTIFIER;
				t->tstrlen = 7;
				t->value = 0;
				locallabel[i]++;
				return 1;
			}
		}
	}
	else if (c == 0x27) {
		/* single quote */
		read1chr();
		t->type = T_SQUOTE;
		return 1;
	}
	else if (c == 0x22) {
		/* double quote */
		read1chr();
		t->type = T_DQUOTE;
		return 1;
	}
	else if (c == '+') {
		/* positive, add */
		read1chr();
		t->type = OPR_PLUS;
		return 1;
	}
	else if (c == '-') {
		/* negative, sub */
		read1chr();
		t->type = OPR_MINUS;
		return 1;
	}
	else if (c == '!') {
		/* not equal, logical not */
		read1chr();
		if (peek1chr() == '=') {
			read1chr();
			t->type = OPR_NE;
			return 1;
		}
		t->type = OPR_NOT;
		return 1;
	}
	else if (c == '~') {
		/* not */
		read1chr();
		t->type = OPR_B_NOT;
		return 1;
	}
	else if (c == '*') {
		/* multiply */
		read1chr();
		t->type = OPR_MUL;
		return 1;
	}
	else if (c == '/') {
		/* division */
		read1chr();
		t->type = OPR_DIV;
		return 1;
	}
	else if (c == '%') {
		/* modulo */
		read1chr();
		t->type = OPR_MOD;
		return 1;
	}
	else if (c == '<') {
		/* less than, less or equal, shift left */
		char c2;
		read1chr();
		c2 = peek1chr();
		if (c2 == '=') {
			read1chr();
			t->type = OPR_LE;
			return 1;
		}
		else if (c2 == '<') {
			read1chr();
			t->type = OPR_SHL;
			return 1;
		}
		t->type = OPR_LT;
		return 1;
	}
	else if (c == '>') {
		/* greater than, greater or equal, shift right */
		char c2;
		read1chr();
		c2 = peek1chr();
		if (c2 == '=') {
			read1chr();
			t->type = OPR_GE;
			return 1;
		}
		else if (c2 == '>') {
			read1chr();
			t->type = OPR_SHR;
			return 1;
		}
		t->type = OPR_GT;
		return 1;
	}
	else if (c == '=') {
		/* equal, re-definable label definition */
		read1chr();
		if (peek1chr() == '=') {
			read1chr();
			t->type = OPR_EQ;
			return 1;
		}
		t->type = CMD_R_EQU;
		return 1;

	}
	else if (c == '&') {
		/* and, logical and */
		read1chr();
		if (peek1chr() == '&') {
			read1chr();
			t->type = OPR_AND;
			return 1;
		}
		t->type = OPR_B_AND;
		return 1;
	}
	else if (c == '|') {
		/* or, logical or */
		read1chr();
		if (peek1chr() == '|') {
			read1chr();
			t->type = OPR_OR;
			return 1;
		}
		t->type = OPR_B_OR;
		return 1;
	}
	else if (c == '^') {
		/* xor */
		read1chr();
		t->type = OPR_B_XOR;
		return 1;
	}
	else if (c == ':') {
		/* label definition */
		read1chr();
		if (peek1chr() == ':') {
			read1chr();
			t->type = CMD_LABELDEF_GLOBAL;
			return 1;
		}
		t->type = CMD_LABELDEF;
		return 1;
	}
	else if (c == '{') {
		/* enter procedure block (namespace) */
		read1chr();
		t->type = CMD_PROC;
		return 1;
	}
	else if (c == '}') {
		/* leave procedure block (namespace) */
		read1chr();
		t->type = CMD_ENDP;
		return 1;
	}
	else if (c == 0) {
		t->type = 0; /*EOS*/
		return 1;
	}
	errmsg = mjrerr[EM_SYNTAX_ERROR];
	return 0;
}
int look_token(token* t) {
	int rv;
	char* oldrp;
	skipspace();
	oldrp = rp;
	rv = get_token(t);
	jp = rp;
	rp = oldrp; /* rewind */
	return rv;
}
int skip_token(void) {
	token dummy;
	if (jp > rp) {
		rp = jp;
		return 1;
	}
	return get_token(&dummy);
}

void get_macro_args(parseinfo* pi) {
	int i;
	char c, * p, * lastp;

	for (i = 0; i < 9; i++) pi->macroargs[i] = 0;
	for (i = 0; i < 9; i++) {
		skipspace();
		p = rp;
		lastp = p;
		while ((c = peek1chr()) != 0) {
			if (c == ',') {
				lastp = rp;
				read1chr();
				break;
			}
			if (isspace(c)) {
				lastp = rp;
				skipspace();
			}
			else
				read1chr();
		}
		if (lastp - p > 0) {
			pi->macroargs[i] = add_linebuf_n(p, lastp - p);
		}
		if (!c) break;
	}
}

/*
 *	putbin
 */
int supress_putbin = 0;
int putbin(unsigned char c) {
	if (supress_putbin) {
		locabs++;
		locabs &= 0x0000FFFF;
		return 1;
	}
	if ((using_in_page && loccnt >= MAXBINBUF * 2) ||
		(!using_in_page && loccnt >= MAXBINBUF)) {
		errmsg = EJ("Binary buffer overflow", "oCiobt@܂");
		return 0;
	}
	if (using_in_page_upto256 && locabs >= 256) {
		errmsg = EJ("IN_PAGE block exceeds 256bytes", "IN_PAGEw͈̔͂256oCg𒴂܂");
		using_in_page_upto256 = 0; // no more duplicate errors
		return 0;
	}
	if (arr_num && locabs >= sav_locabs) {
		errmsg = EJ("ARRANGE overflow", "ARRANGEł̔zumۍςݗ̈𒴂܂");
		return 0;
	}
	/*dbg printf("%04X:%04X = %02X%s\n",loccnt,locabs,c, (!pinfo->pass2) ? "" : " (pass 2)");*/
	if (!pinfo->pass2) binbuf[loccnt] = c;
	loccnt++;
	locabs++;
	if (arr_num) {
		arr_count++;
		if (arr_count >= arr_bytes[arr_index]) {
			arr_count = 0;
			arr_index++;
			locabs += arr_bytes[arr_index];
			loccnt += arr_bytes[arr_index];
			arr_index++;
			if (arr_index >= arr_num) arr_index = 0;
		}
	}
	return 1;
}

static small_hira = 0;
int putbin_transcode(unsigned char c) {
	int i;
	i = (int)(c & 0x000000ff);
	switch (db_encode) {
	case DB_ASCII_MZ_JP:
		if (i >= 'A' && i <= 'Z') {
			if (small_hira) {
				if (!putbin(0x05)) return 0;
				small_hira = 0;
			}
		}
		else if (i >= 'a' && i <= 'z') {
			if (!small_hira) {
				if (!putbin(0x05)) return 0;
				small_hira = 1;
			}
			c = (unsigned char)(i - ('a' - 'A'));
		}
		else if (i >= 0xA1 && i <= 0xDF) { // han-kana
			c = (unsigned char)(i - 0x20); // to MZ-kana
		}
		break;

	case DB_ASCII_MZ_EU:
		if (i >= 0x60 && i <= 0x7E) {
			c = euasc[i - 0x60];
		}
		else if (i == 0x5E) { // '^'
			c = 0x8B;
		}
		else if (i == 0x5F) { // '_'
			c = 0x90;
		}
		break;

	case DB_DISP_JP:
		if (i >= 0x20 && i <= 0x7f) c = dispc[i - 0x20];
		else if (i >= 0xa0 && i <= 0xdf) c = dispck[i - 0xa0];
		break;

	case DB_DISP_EU:
		if (i >= 0x20 && i <= 0x7f) c = dispc_eu[i - 0x20];
		break;

	case DB_USER_DEFINED:
		c = mapchr[i];
		break;

	case DB_ASCII_GENERIC: // ASCII
	default:
		break;
	}
	return putbin(c);
}

//
// Checks if a special binary placement mode is running / ȃoCizu[hsǂ`FbN
//
int check_special_putbin(char* dstr) {
	static char* cstr;
	if (supress_putbin || arr_num || using_in_page) {
		sprintf(emstr, EJ("%s can't be used in %s", "%sw蒆%s͎gpł܂"), EJ(dstr, cstr), EJ(cstr, dstr));
		errmsg = emstr;
		return 0;
	}
	cstr = dstr;
	return 1;
}

/* -------------------------------------- */

/*
 *	Calc Expression
 */
static long expr_vstack[256];	/* value stack */
static int expr_tstack[256];	/* type stack */
static int expr_sp;
/* following 3 global variables are for the ambigious addressing expression
   of Z80 assembly language... sigh */
static int expr_end_w_brc;	/* 1 if last token is a closing brace */
static int expr_brc_nest;	/* nest level of braces (0:ground level) */
static int expr_ex_brc_cnt;	/* how many times braces closed at the ground level
				   it is necessary to find '(1+2)*(3+4)' types */
void expr_push(long v, int typ) {
	expr_vstack[expr_sp] = v;
	expr_tstack[expr_sp] = typ;
	expr_sp++;
}
void expr_pop(long* v, int* typ) {
	--expr_sp;
	*v = expr_vstack[expr_sp];
	*typ = expr_tstack[expr_sp];
}
unres_expr* extract_expr(int isrel) {
	unres_expr* p;
	long* vs;
	int* ts;
	int i, s;
	s = isrel ? 2 : 0; /* if relative-jump, extract as "<expr> - <baseaddr>" */
	if (isrel && using_in_page) s += 2; /* if using in_page, extract as "<expr> - (<label> + <offset>)" */
	p = (unres_expr*)malloc(sizeof(unres_expr) + (sizeof(long) + sizeof(int)) * (expr_sp + s));
	p->depth = expr_sp + s;
	vs = (long*)&(p->vstack[0]);
	for (i = 0; i < expr_sp; i++) *vs++ = expr_vstack[i];
	if (isrel) {
		if (!using_in_page) {
			*vs++ = locabs + 1;     // "<expr> - <baseaddr>"
			*vs++ = 0;
		}
		else {
			*vs++ = using_in_page;  // "<expr> - (<label> + <offset>)"
			*vs++ = locabs + 1;
			*vs++ = 0;
			*vs++ = 0;
		}
	}
	ts = (int*)vs;
	for (i = 0; i < expr_sp; i++) *ts++ = expr_tstack[i];
	if (isrel) {
		if (!using_in_page) {
			*ts++ = IMMIDIATE;
			*ts = OPR_MINUS;
		}
		else {
			*ts++ = IDENTIFIER;
			*ts++ = IMMIDIATE;
			*ts++ = OPR_PLUS;
			*ts = OPR_MINUS;
		}
	}
	return p;
}
/*	expr0 -	immidiate
		identifier
		'(' expr12 ')'	*/
int expr12(void);
int expr0(void) {
	long li;
	label* l;
	token t;
	if (!look_token(&t)) return 0;
	expr_end_w_brc = 0;
	switch (t.type) {
	case IMMIDIATE:
		expr_push(t.value, IMMIDIATE);
		skip_token();
		break;
	case IDENTIFIER:
		li = getlabel(t.tstr);
		if (!li) {
			li = entrylabel(t.tstr);
			if (!li) return 0;
			l = getlabelptr(li);
			l->value = 0;
			l->defstate = LABEL_UNDEFINED;
		}
		l = getlabelptr(li);
		if (l->defstate == LABEL_DEFINED)
			expr_push(l->value, IMMIDIATE);
		else
			expr_push(li, IDENTIFIER);
		skip_token();
		break;
	case T_BRCOPEN:
		expr_brc_nest++;
		skip_token();
		if (!expr12()) return 0;
		if (!get_token(&t)) return 0;
		if (t.type != T_BRCCLOSE) {
			errmsg = mjrerr[EM_MISSING_RBRC];
			return 0;
		}
		expr_end_w_brc = 1;
		expr_brc_nest--;
		if (expr_brc_nest == 0) expr_ex_brc_cnt++;
		break;
	case T_SQUOTE:
		skip_token();
		li = 0;
		while (peek1chr() != 0x27 && peek1chr() != 0) {
			li <<= 8;
			li |= ((long)read1chr()) & 0x000000ff;
		}
		if (read1chr() != 0x27) {
			errmsg = EJ("Missing closing '''", "I'''܂");
			return 0;
		}
		expr_push(li, IMMIDIATE);
		break;
	case OPR_DEFINED:
		skip_token();
		if (!get_token(&t) || t.type != T_BRCOPEN) {
			errmsg = "Missing '('";
			return 0;
		}
		if (!get_token(&t) || t.type != IDENTIFIER) {
			errmsg = "Illegal label in DEFINED()";
			return 0;
		}
		li = getlabel(t.tstr);
		if (pinfo->pass2) {
			expr_push(li && ((getlabelptr(li))->reftype & LF_APPEARED_IN_PASS2), IMMIDIATE);
		}
		else {
			expr_push(li && (getlabelptr(li))->defstate != LABEL_UNDEFINED, IMMIDIATE);
		}
		if (!get_token(&t) || t.type != T_BRCCLOSE) {
			errmsg = mjrerr[EM_MISSING_RBRC];
			return 0;
		}
		break;
	case T_LOCCNT:
		skip_token();
		if (!using_in_page) {
			expr_push(locdol, IMMIDIATE);
		}
		else {
			expr_push(using_in_page, IDENTIFIER);
			expr_push(locdol, IMMIDIATE);
			expr_push(0, OPR_PLUS);
		}
		break;
	default:
		errmsg = EJ("Syntax error in expression", "̍\G[");
		return 0;
	}
	return 1;
}
/*	expr1 -	['+'] expr0
		['-'] expr0
		['!'] expr0
		['~'] expr0	*/
int expr1(void) {
	token t;
	long v;
	int d;
	if (!look_token(&t)) return 0;
	switch (t.type) {
	case OPR_PLUS:
		skip_token();
		return expr1();
	case OPR_MINUS:
		skip_token();
		if (!expr1()) return 0;
		if (expr_tstack[expr_sp - 1] == IMMIDIATE) {
			expr_pop(&v, &d);
			expr_push(-v, IMMIDIATE);
		}
		else
			expr_push(0, OPR_NEG);
		return 1;
	case OPR_NOT:
		skip_token();
		if (!expr1()) return 0;
		if (expr_tstack[expr_sp - 1] == IMMIDIATE) {
			expr_pop(&v, &d);
			expr_push(!v, IMMIDIATE);
		}
		else
			expr_push(0, OPR_NOT);
		return 1;
	case OPR_B_NOT:
		skip_token();
		if (!expr1()) return 0;
		if (expr_tstack[expr_sp - 1] == IMMIDIATE) {
			expr_pop(&v, &d);
			expr_push(~v, IMMIDIATE);
		}
		else
			expr_push(0, OPR_B_NOT);
		return 1;
	}
	return expr0();
}
/*	expr2 -	expr1 ['*' expr1]
		expr1 ['/' expr1]
		expr1 ['%' expr1]	*/
int expr2(void) {
	token t;
	long v1, v2;
	int d;
	if (!expr1()) return 0;
	while (look_token(&t) &&
		(t.type == OPR_MUL ||
			t.type == OPR_DIV ||
			t.type == OPR_MOD)) {
		skip_token();
		if (!expr1()) return 0;
		if (expr_tstack[expr_sp - 1] == IMMIDIATE &&
			expr_tstack[expr_sp - 2] == IMMIDIATE) {
			switch (t.type) {
			case OPR_MUL:
				expr_pop(&v2, &d);
				expr_pop(&v1, &d);
				expr_push(v1 * v2, IMMIDIATE);
				break;
			case OPR_DIV:
				expr_pop(&v2, &d);
				expr_pop(&v1, &d);
				expr_push(v1 / v2, IMMIDIATE);
				break;
			case OPR_MOD:
				expr_pop(&v2, &d);
				expr_pop(&v1, &d);
				expr_push(v1 % v2, IMMIDIATE);
				break;
			}
		}
		else
			expr_push(0, t.type);
	}
	return 1;
}
/*	expr3 -	expr2 ['+' expr2]
		expr2 ['-' expr2]	*/
int expr3(void) {
	token t;
	long v1, v2;
	int d;
	if (!expr2()) return 0;
	while (look_token(&t) &&
		(t.type == OPR_PLUS ||
			t.type == OPR_MINUS)) {
		skip_token();
		if (!expr2()) return 0;
		if (expr_tstack[expr_sp - 1] == IMMIDIATE &&
			expr_tstack[expr_sp - 2] == IMMIDIATE) {
			if (t.type == OPR_PLUS) {
				expr_pop(&v2, &d);
				expr_pop(&v1, &d);
				expr_push(v1 + v2, IMMIDIATE);
			}
			else {
				expr_pop(&v2, &d);
				expr_pop(&v1, &d);
				expr_push(v1 - v2, IMMIDIATE);
			}
		}
		else
			expr_push(0, t.type);
	}
	return 1;
}
/*	expr4 -	expr3 ['<<' expr3]
		expr3 ['>>' expr3]	*/
int expr4(void) {
	token t;
	long v1, v2;
	int d;
	if (!expr3()) return 0;
	while (look_token(&t) &&
		(t.type == OPR_SHL ||
			t.type == OPR_SHR)) {
		skip_token();
		if (!expr3()) return 0;
		if (expr_tstack[expr_sp - 1] == IMMIDIATE &&
			expr_tstack[expr_sp - 2] == IMMIDIATE) {
			if (t.type == OPR_SHL) {
				expr_pop(&v2, &d);
				expr_pop(&v1, &d);
				expr_push(v1 << v2, IMMIDIATE);
			}
			else {
				expr_pop(&v2, &d);
				expr_pop(&v1, &d);
				expr_push(v1 >> v2, IMMIDIATE);
			}
		}
		else
			expr_push(0, t.type);
	}
	return 1;
}
/*	expr5 -	expr4 ['>' expr4]
		expr4 ['<' expr4]
		expr4 ['>=' expr4]
		expr4 ['<=' expr4]	*/
int expr5(void) {
	token t;
	long v1, v2;
	int d;
	if (!expr4()) return 0;
	while (look_token(&t) &&
		(t.type == OPR_GT ||
			t.type == OPR_LT ||
			t.type == OPR_GE ||
			t.type == OPR_LE)) {
		skip_token();
		if (!expr4()) return 0;
		if (expr_tstack[expr_sp - 1] == IMMIDIATE &&
			expr_tstack[expr_sp - 2] == IMMIDIATE) {
			switch (t.type) {
			case OPR_GT:
				expr_pop(&v2, &d);
				expr_pop(&v1, &d);
				expr_push(v1 > v2, IMMIDIATE);
				break;
			case OPR_LT:
				expr_pop(&v2, &d);
				expr_pop(&v1, &d);
				expr_push(v1 < v2, IMMIDIATE);
				break;
			case OPR_GE:
				expr_pop(&v2, &d);
				expr_pop(&v1, &d);
				expr_push(v1 >= v2, IMMIDIATE);
				break;
			case OPR_LE:
				expr_pop(&v2, &d);
				expr_pop(&v1, &d);
				expr_push(v1 <= v2, IMMIDIATE);
				break;
			}
		}
		else
			expr_push(0, t.type);
	}
	return 1;
}
/*	expr6 -	expr5 ['==' expr5]
		expr5 ['!=' expr5]	*/
int expr6(void) {
	token t;
	long v1, v2;
	int d;
	if (!expr5()) return 0;
	while (look_token(&t) &&
		(t.type == OPR_EQ ||
			t.type == OPR_NE)) {
		skip_token();
		if (!expr5()) return 0;
		if (expr_tstack[expr_sp - 1] == IMMIDIATE &&
			expr_tstack[expr_sp - 2] == IMMIDIATE) {
			if (t.type == OPR_EQ) {
				expr_pop(&v2, &d);
				expr_pop(&v1, &d);
				expr_push(v1 == v2, IMMIDIATE);
			}
			else {
				expr_pop(&v2, &d);
				expr_pop(&v1, &d);
				expr_push(v1 != v2, IMMIDIATE);
			}
		}
		else
			expr_push(0, t.type);
	}
	return 1;
}
/*	expr7 -	expr6 ['&' expr6]	*/
int expr7(void) {
	token t;
	long v1, v2;
	int d;
	if (!expr6()) return 0;
	while (look_token(&t) &&
		t.type == OPR_B_AND) {
		skip_token();
		if (!expr6()) return 0;
		if (expr_tstack[expr_sp - 1] == IMMIDIATE &&
			expr_tstack[expr_sp - 2] == IMMIDIATE) {
			expr_pop(&v2, &d);
			expr_pop(&v1, &d);
			expr_push(v1 & v2, IMMIDIATE);
		}
		else
			expr_push(0, t.type);
	}
	return 1;
}
/*	expr8 -	expr7 ['^' expr7]	*/
int expr8(void) {
	token t;
	long v1, v2;
	int d;
	if (!expr7()) return 0;
	while (look_token(&t) &&
		t.type == OPR_B_XOR) {
		skip_token();
		if (!expr7()) return 0;
		if (expr_tstack[expr_sp - 1] == IMMIDIATE &&
			expr_tstack[expr_sp - 2] == IMMIDIATE) {
			expr_pop(&v2, &d);
			expr_pop(&v1, &d);
			expr_push(v1 ^ v2, IMMIDIATE);
		}
		else
			expr_push(0, t.type);
	}
	return 1;
}
/*	expr9 -	expr8 ['|' expr8]	*/
int expr9(void) {
	token t;
	long v1, v2;
	int d;
	if (!expr8()) return 0;
	while (look_token(&t) &&
		t.type == OPR_B_OR) {
		skip_token();
		if (!expr8()) return 0;
		if (expr_tstack[expr_sp - 1] == IMMIDIATE &&
			expr_tstack[expr_sp - 2] == IMMIDIATE) {
			expr_pop(&v2, &d);
			expr_pop(&v1, &d);
			expr_push(v1 | v2, IMMIDIATE);
		}
		else
			expr_push(0, t.type);
	}
	return 1;
}
/*	expr10 - expr9 ['&&' expr9]	*/
int expr10(void) {
	token t;
	long v1, v2;
	int d;
	if (!expr9()) return 0;
	while (look_token(&t) &&
		t.type == OPR_AND) {
		skip_token();
		if (!expr9()) return 0;
		if (expr_tstack[expr_sp - 1] == IMMIDIATE &&
			expr_tstack[expr_sp - 2] == IMMIDIATE) {
			expr_pop(&v2, &d);
			expr_pop(&v1, &d);
			expr_push(v1 && v2, IMMIDIATE);
		}
		else
			expr_push(0, t.type);
	}
	return 1;
}
/*	expr11 - expr10 ['||' expr10]	*/
int expr11(void) {
	token t;
	long v1, v2;
	int d;
	if (!expr10()) return 0;
	while (look_token(&t) &&
		t.type == OPR_OR) {
		skip_token();
		if (!expr10()) return 0;
		if (expr_tstack[expr_sp - 1] == IMMIDIATE &&
			expr_tstack[expr_sp - 2] == IMMIDIATE) {
			expr_pop(&v2, &d);
			expr_pop(&v1, &d);
			expr_push(v1 || v2, IMMIDIATE);
		}
		else
			expr_push(0, t.type);
	}
	return 1;
}
/*	expr12 - expr11 [':' expr11]	*/
int expr12(void) {
	token t;
	long v1, v2;
	int d;
	if (!expr11()) return 0;
	while (look_token(&t) &&
		t.type == OPR_HL) {
		skip_token();
		if (!expr11()) return 0;
		if (expr_tstack[expr_sp - 1] == IMMIDIATE &&
			expr_tstack[expr_sp - 2] == IMMIDIATE) {
			expr_pop(&v2, &d);
			expr_pop(&v1, &d);
			expr_push((v1 << 8) | (v2 & 0x00FF), IMMIDIATE);
		}
		else
			expr_push(0, t.type);
	}
	return 1;
}
static int expr_braced;
static int expr_finished;
int expression(void) {
	token t;
	int r;
	int expr_start_w_brc = 0;
	expr_brc_nest = expr_ex_brc_cnt = 0;
	expr_sp = 0;
	errmsg = mjrerr[EM_MISSING_RBRC];
	if (look_token(&t) && t.type == T_BRCOPEN) {
		expr_start_w_brc = 1;
	}
	r = expr12();
	expr_braced = (expr_start_w_brc && expr_end_w_brc && expr_ex_brc_cnt == 1);
	expr_finished = (expr_sp == 1 && expr_tstack[0] == IMMIDIATE);
	if (r) errmsg = 0;
	return r;
}
int is_expr_braced(void) { return expr_braced; }
int is_expr_finished(void) { return expr_finished; }
long expr_result(void) { return expr_vstack[0]; }

int expr_label_with_offset(long li, int offset) {
	expr_sp = 0;
	expr_push(li, IDENTIFIER);
	expr_push(offset, IMMIDIATE);
	expr_push(0, OPR_PLUS);
	return 1;
}

/* -------------------------------------- */

/*
 *	unresolved expression info
 */
static unres_expr* unres_expr_chain, * unres_expr_chain_tail;
void entry_unres_label(int loc, int typ, int lnum) {
	unres_expr* p;
	p = extract_expr(typ == PATCH_RELJMP);
	p->next = 0;
	p->patchloc = loc;
	p->patchtype = typ;
	p->linenum = lnum;
	p->filename = pinfo->filename;
	p->namespaceid = get_namespace_id();
	if (!unres_expr_chain) {
		/* first entry */
		unres_expr_chain = p;
	}
	else {
		unres_expr_chain_tail->next = p;
	}
	unres_expr_chain_tail = p;
}
/* proc unresolved expression */
int patch_unres_expr(void) {
	int retval = 1;
	int i;
	int skipthis, proccnt;
	unres_expr* p, * p0, ** pl;
	long* vs;
	int* ts;
	int sp;
	long stk[256];
	long v;
	label* l;
	proccnt = -1; /*dummy*/
/*DBG for(p=unres_expr_chain; p; p=p->next){
  switch(p->patchtype) {
	case 0: printf("Label: %s\n", getlabelstr(p->patchloc)); break;
	case 1: printf("Byte : %04X\n", p->patchloc); break;
	case 2: printf("Word : %04X\n", p->patchloc); break;
	case 3: printf("RelJP: %04X\n", p->patchloc); break;
	case 4: printf("BitCt: %04X\n", p->patchloc); break;
  }
}*/
	while (unres_expr_chain && proccnt != 0) {
		p = unres_expr_chain;
		pl = &unres_expr_chain;
		proccnt = 0;
		while (p) {
			skipthis = 0;
			vs = (long*)&(p->vstack[0]);
			ts = (int*)&vs[p->depth];
			sp = 0;
			switch_namespace(p->namespaceid);
			for (i = 0; i < p->depth; i++) {
				switch (ts[i]) {
				case IMMIDIATE:
					stk[sp++] = vs[i];
					break;
				case IDENTIFIER:
					l = getlabelptr(vs[i]);
					if (l->defstate == LABEL_DEFINED) {
						stk[sp++] = l->value;
					}
					else {
						/* search upper namespaces */
						char buf[MAXTOKENLEN];
						label* l0;
						strcpy(buf, getlabelstr(vs[i]));
						if (buf[0] != '@' &&	// local label cannot go out from scope
							(l0 = getlabelptr_outside(buf, LABEL_DEFINED)) != 0) {
							stk[sp++] = l0->value;
							/* copy the value for pass-2 */
							l->defstate = LABEL_DEFINED;
							l->value = l0->value;
							l->reftype |= LF_HIDE;	// do not show it
						}
						else {
							skipthis = 1;
						}
					}
					break;
				case OPR_NEG:
					v = stk[--sp];
					stk[sp++] = -v;
					break;
				case OPR_NOT:
					v = stk[--sp];
					stk[sp++] = !v;
					break;
				case OPR_B_NOT:
					v = stk[--sp];
					stk[sp++] = ~v;
					break;
				case OPR_PLUS:
					v = stk[--sp];
					v += stk[--sp];
					stk[sp++] = v;
					break;
				case OPR_MINUS:
					v = stk[--sp];
					v = stk[--sp] - v;
					stk[sp++] = v;
					break;
				case OPR_MUL:
					v = stk[--sp];
					v *= stk[--sp];
					stk[sp++] = v;
					break;
				case OPR_DIV:
					v = stk[--sp];
					v = stk[--sp] / v;
					stk[sp++] = v;
					break;
				case OPR_MOD:
					v = stk[--sp];
					v = stk[--sp] % v;
					stk[sp++] = v;
					break;
				case OPR_SHL:
					v = stk[--sp];
					v = stk[--sp] << v;
					stk[sp++] = v;
					break;
				case OPR_SHR:
					v = stk[--sp];
					v = stk[--sp] >> v;
					stk[sp++] = v;
					break;
				case OPR_GT:
					v = stk[--sp];
					v = stk[--sp] > v;
					stk[sp++] = v;
					break;
				case OPR_LT:
					v = stk[--sp];
					v = stk[--sp] < v;
					stk[sp++] = v;
					break;
				case OPR_GE:
					v = stk[--sp];
					v = stk[--sp] >= v;
					stk[sp++] = v;
					break;
				case OPR_LE:
					v = stk[--sp];
					v = stk[--sp] <= v;
					stk[sp++] = v;
				case OPR_EQ:
					v = stk[--sp];
					v = stk[--sp] == v;
					stk[sp++] = v;
					break;
				case OPR_NE:
					v = stk[--sp];
					v = stk[--sp] != v;
					stk[sp++] = v;
				case OPR_B_AND:
					v = stk[--sp];
					v &= stk[--sp];
					stk[sp++] = v;
					break;
				case OPR_B_OR:
					v = stk[--sp];
					v |= stk[--sp];
					stk[sp++] = v;
					break;
				case OPR_B_XOR:
					v = stk[--sp];
					v ^= stk[--sp];
					stk[sp++] = v;
					break;
				case OPR_AND:
					v = stk[--sp];
					v = stk[--sp] && v;
					stk[sp++] = v;
					break;
				case OPR_OR:
					v = stk[--sp];
					v = stk[--sp] || v;
					stk[sp++] = v;
					break;
				case OPR_HL:
					v = stk[--sp];
					v = (stk[--sp] << 8) | (v & 0x00FF);
					stk[sp++] = v;
					break;
				default:
					printf("Error %s%5d: Unexpected expression mid-code\n",
						get_string(p0->filename), p0->linenum);
					return 0;
				}
				if (skipthis) break;
			}
			p0 = p;
			p = p->next;
			if (!skipthis) {
				*pl = p;
				proccnt++;
				/* patch */
				v = stk[--sp];
				switch (p0->patchtype) {
				case PATCH_BYTE: /* byte */
					binbuf[p0->patchloc] = (unsigned char)(v & 0xff);
					break;
				case PATCH_WORD: /* word */
					binbuf[p0->patchloc] = (unsigned char)(v & 0xff);
					binbuf[p0->patchloc + 1] = (unsigned char)((v >> 8) & 0xff);
					break;
				case PATCH_RELJMP: /* special, relative jump */
					if (v > 127 || v < -128) {
						printf("Error %s%5d: %s (%+d)\n",
							get_string(p0->filename), p0->linenum, mjrerr[EM_JR_OVERFLOW],
							(v > 127) ? v - 127 : v + 128);
						retval = 0;
					}
					else {
						binbuf[p0->patchloc] = (unsigned char)(v & 0xff);
					}
					break;
				case PATCH_BITCNT: /* special, bit count */
					if (v > 7 || v < 0) {
						printf("Error %s%5d: %s\n", get_string(p0->filename), p0->linenum, mjrerr[EM_BITCNT_OVERFLOW]);
						retval = 0;
					}
					else {
						binbuf[p0->patchloc] |= (unsigned char)(v << 3);
					}
					break;
				case PATCH_LABEL: /* label value */
					l = getlabelptr(p0->patchloc);
					l->value = (unsigned short)(v & 0x0000ffff);
					l->defstate = LABEL_DEFINED;
				}
				/* dispose it */
				free(p0);
			}
			else {
				pl = &(p0->next);
			}
		}
	}
	while (unres_expr_chain) {
		p = unres_expr_chain;
		vs = (long*)&(p->vstack[0]);
		ts = (int*)&(vs[p->depth]);
		for (i = 0; i < p->depth; i++) {
			if (ts[i] == IDENTIFIER) {
				char buf[256];
				l = getlabelptr(vs[i]);
				if (l->defstate == LABEL_UNRESSHOWN) continue;
				if ((l->reftype & (LF_PAGE_ALIGNED | LF_IN_PAGE)) != 0) continue; // it's not an error by a user
				strcpy(buf, getlabelstr(vs[i]));
				if (buf[0] == '@') {
					buf[2] = 0;
				}
				printf(EJ("Error %s%5d: Unresolved label \'%s\'\n", "Error %s%5d: ̃x \'%s\'\n"),
					get_string(p->filename), p->linenum, buf);
				l->defstate = LABEL_UNRESSHOWN;
				retval = 0;
			}
		}
		p0 = p->next;
		free(p);
		unres_expr_chain = p0;
	}

	return retval;
}
void dispose_unres_expr(void) {
	unres_expr* p;
	while (unres_expr_chain) {
		p = unres_expr_chain;
		unres_expr_chain = p->next;
		free(p);
	}
}

/* -------------------------------------- */

int get_comma(void) {
	token t;
	if (!look_token(&t)) return 0;
	if (t.type != T_COMMA) {
		errmsg = EJ("Missing ','", "J}','܂");
		return 0;
	}
	skip_token();
	return 1;
}

int chk_reg8(int typ, unsigned char* pfx) {
	*pfx = 0;
	switch (typ) {
	case REG_A: return 7;
	case REG_B: return 0;
	case REG_C: return 1;
	case REG_D: return 2;
	case REG_E: return 3;
	case REG_H: return 4;
	case REG_L: return 5;
	case REG_IXH: *pfx = 0xDD; return 4;
	case REG_IXL: *pfx = 0xDD; return 5;
	case REG_IYH: *pfx = 0xFD; return 4;
	case REG_IYL: *pfx = 0xFD; return 5;
	case REG_HL_BRACED: return 6;
	case REG_IX_BRACED: *pfx = 0xDD; return 6;
	case REG_IY_BRACED: *pfx = 0xFD; return 6;
	case REG_IX_d_BRACED: *pfx = 0xDD; return 6;
	case REG_IY_d_BRACED: *pfx = 0xFD; return 6;
	}
	return -1;
}

int chk_reg16(int typ, unsigned char* pfx) {
	*pfx = 0;
	switch (typ) {
	case REG_BC: return 0;
	case REG_DE: return 1;
	case REG_HL: return 2;
	case REG_SP: return 3;
	case REG_IX: *pfx = 0xDD; return 2;
	case REG_IY: *pfx = 0xFD; return 2;
	}
	return -1;
}

int chk_reg16_1(int typ, unsigned char pfx) {	/* pfx -> 0:HL 0xDD:IX 0xFD:IY */
	switch (typ) {
	case REG_BC: return 0;
	case REG_DE: return 1;
	case REG_HL: return (pfx == 0) ? 2 : -1;
	case REG_SP: return 3;
	case REG_IX: return (pfx == 0xDD) ? 2 : -1;
	case REG_IY: return (pfx == 0xFD) ? 2 : -1;
	}
	return -1;
}
int chk_reg16_2(int typ, unsigned char* pfx) {
	*pfx = 0;
	switch (typ) {
	case REG_BC: return 0;
	case REG_DE: return 1;
	case REG_HL: return 2;
	case REG_AF: return 3;
	case REG_IX: *pfx = 0xDD; return 2;
	case REG_IY: *pfx = 0xFD; return 2;
	}
	return -1;
}

int chk_cond4(int typ) {
	switch (typ) {
	case CND_NZ: return 0;
	case CND_Z: return 1;
	case CND_NC: return 2;
	case CND_C: return 3;
	}
	return -1;
}

int chk_cond8(int typ) {
	switch (typ) {
	case CND_NZ: return 0;
	case CND_Z: return 1;
	case CND_NC: return 2;
	case CND_C: return 3;
	case CND_PO: return 4;
	case CND_PE: return 5;
	case CND_P: return 6;
	case CND_M: return 7;
	}
	return -1;
}

int relative_address(void) {
	int e;
	if (!expression()) return 0;
	if (is_expr_finished()) {
		e = expr_result() - (locabs + 1);
		if (e > 127 || e < -128) {
			sprintf(emstr, "%s (%+d)", mjrerr[EM_JR_OVERFLOW],
				(e > 127) ? e - 127 : e + 128);
			errmsg = emstr;
			return 0;
		}
		if (!putbin((unsigned char)e)) return 0;
	}
	else {
		entry_unres_label(loccnt, PATCH_RELJMP, pinfo->linenum);
		if (!putbin(0)) return 0;
	}
	return 1;
}

int imm8(void) {
	if (!expression()) return 0;
	if (is_expr_finished()) {
		if (!putbin((unsigned char)(expr_result() & 0xff))) return 0;
	}
	else {
		entry_unres_label(loccnt, PATCH_BYTE, pinfo->linenum);
		if (!putbin(0)) return 0;
	}
	return 1;
}

int imm16(void) {
	long nn;
	if (!expression()) return 0;
	if (is_expr_finished()) {
		nn = expr_result();
	}
	else {
		entry_unres_label(loccnt, PATCH_WORD, pinfo->linenum);
		nn = 0;
	}
	if (!putbin((unsigned char)(nn & 0xff))) return 0;
	if (!putbin((unsigned char)((nn >> 8) & 0xff))) return 0;
	return 1;
}

int reg8_(unsigned char opc, unsigned char cbi) {
	token t;
	unsigned char pfx;
	int i;
	if (!look_token(&t)) return 0;
	/* r */
	if ((i = chk_reg8(t.type, &pfx)) >= 0) {
		skip_token();
		if (pfx) {
			if (!putbin(pfx)) return 0;
		}

		/* Check for DD CB, FD CB undefined instructions */
		if (cbi && (opc & 0x07) != 6) {
			if (i != 6 && !pfx) return 0;
		}
		else {
			opc &= 0xF8;
			opc |= i;
		}

		/* CB instructions: DD CB d opc */
		if (!putbin(cbi ? cbi : opc)) return 0;

		if (t.type == REG_IX_d_BRACED || t.type == REG_IY_d_BRACED) {
			if (!imm8()) return 0;
			if (!get_token(&t) || t.type != T_BRCCLOSE) return 0;
		}
		else if (t.type == REG_IX_BRACED || t.type == REG_IY_BRACED) {
			if (!putbin(0)) return 0;
		}
		/* regular instructions: DD opc d */
		if (cbi) {
			if (!putbin(opc)) return 0;
		}
	}
	else return 2; /* no error, maybe imm8 */
	return 1;
}

int reg8(unsigned char opc, unsigned char cbi) {
	return (reg8_(opc, cbi) == 1) ? 1 : 0;
}

int arithmetic8(unsigned char opcr, unsigned char opcn) {
	switch (reg8_(opcr, 0)) {
	case 1: return 1;
	case 2:
		if (!putbin(opcn)) return 0;
		return imm8();
	}
	return 0;
}

void skip_A_comma(void) {
	token t;
	char* oldrp;
	oldrp = rp;
	if (get_token(&t) && t.type == REG_A) {
		if (get_token(&t) && t.type == T_COMMA) {
			printf(EJ("Warning %s %d: Redundant operand 'A,'\n",
				"Warning %s %d: 璷ȃIyh 'A,'\n"), get_string(pinfo->filename), pinfo->linenum);
			return;
		}
	}
	rp = oldrp; /* rewind */
}

//
// label definition
//
long define_label(char* lstr, int is_across_namespace) {
	label* l;
	long li;
	if (!(li = getlabel(lstr))) {
		// if not defined yet, try to entry the label
		if (!is_across_namespace) {
			li = entrylabel(lstr);
		}
		else {
			li = entrylabel_across_namespace(lstr, 0);
		}
		if (!li) return 0;
	}
	else {
		// if already defined, check if it is just formerly referred
		if (is_across_namespace) {
			li = entrylabel_across_namespace(NULL, li);
		}
		l = getlabelptr(li);
		if (l->defstate != LABEL_UNDEFINED) {
			em_dup_label(lstr, l->defstate);
			return 0;
		}
	}
	return li;
}

int define_location_label(long li) {
	label* l;
	if (using_in_page) {
		expr_label_with_offset(using_in_page, locabs);
		entry_unres_label(li, PATCH_LABEL, pinfo->linenum);
	}
	else {
		/* just define the current location */
		l = getlabelptr(li);
		l->value = locabs;
		l->defstate = LABEL_DEFINED;
	}
	return 1;
}

void init_using_in_page() {
	using_in_page = 0;
	using_in_page_type = 0;
	using_in_page_upto256 = 0;
	using_in_page_used = 0;
	loccnt_in_page = MAXBINBUF;
}
void finish_using_in_page() {
	label* l;
	if (pinfo->pass2) goto xit;
	if (!using_in_page) return;
	l = getlabelptr(using_in_page);
	l->value2 = locabs; // size of this block
	using_in_page = 0;
	loccnt_in_page = loccnt;
xit:
	using_in_page_upto256 = 0;
	using_in_page_type = 0;
	locabs = sav_locabs;
	loccnt = sav_loccnt;
	locdol = locabs;
}

static int loccnt_s, locabs_s, loccnt_e;
int parse(char* linebuf, parseinfo* pi) {
	int i;
	int didcut, doflush;
	char outasm[256], * copyfrom, * copyto, * wp;
	token t;
	int labelval_pass2;
	rp = linebuf;
	errmsg = 0;
	locdol = locabs;
	pinfo = pi;
#ifdef _DEBUG
	fprintf(stdout, "%4d: %s", pi->linenum, linebuf);
#endif // _DEBUG
	if (pi->pass2) {
		loccnt_s = loccnt;
		locabs_s = locabs;
		copyfrom = rp;
		wp = outasm;
		*wp = 0;
		didcut = 0;
		doflush = 0;
	}
loop:
	labelval_pass2 = 0;
	while (errmsg == 0 && get_token(&t)/* && t.type != 0*/) {
		copyto = 0;
		if (INST_START <= t.type && t.type <= INST_END) {
			unsigned char pfx, opc1, opc2;
			long nn;
			pfx = opc1 = opc2 = 0;
			switch (t.type) {

				/* ADD A,r : ADD A,n : ADD HL,ss */
			case INST_ADD:
				if (!look_token(&t)) break;
				if (t.type == REG_A) {
					skip_token();
					if (!get_comma()) break;
					/* A,(HL) : A,(IX+d) : A,(IY+d) */
					/* A,r : A,n */
					if (!arithmetic8(0x80, 0xc6)) errmsg = mjrerr[EM_ILLEGAL_OPERAND];
				}
				else if (t.type == REG_HL || t.type == REG_IX || t.type == REG_IY) {
					int ss;
					skip_token();
					if (t.type == REG_IX) pfx = 0xDD;
					else if (t.type == REG_IY) pfx = 0xFD;
					if (pfx && !putbin(pfx)) break;
					if (!get_comma()) break;
					if (!look_token(&t)) break;
					ss = chk_reg16_1(t.type, pfx);
					if (ss < 0) {
						errmsg = mjrerr[EM_ILLEGAL_OPERAND];
						break;
					}
					skip_token();
					/* 00 ss1 001 */
					if (!putbin((unsigned char)(0x09 | (ss << 4)))) break;
				}
				else
					errmsg = mjrerr[EM_ILLEGAL_OPERAND];
				break;

				/* ADC A,r : ADC A,n : ADC HL,ss */
			case INST_ADC:
				if (!look_token(&t)) break;
				if (t.type == REG_A) {
					skip_token();
					if (!get_comma()) break;
					/* A,(HL) : A,(IX+d) : A,(IY+d) */
					/* A,r : A,n */
					if (!arithmetic8(0x88, 0xce)) errmsg = mjrerr[EM_ILLEGAL_OPERAND];
				}
				else if (t.type == REG_HL) {
					int ss;
					skip_token();
					if (!get_comma()) break;
					if (!look_token(&t)) break;
					ss = chk_reg16_1(t.type, 0);
					if (ss < 0) {
						errmsg = mjrerr[EM_ILLEGAL_OPERAND];
						break;
					}
					skip_token();
					/* ED : 01 ss1 010 */
					if (!putbin((unsigned char)(0xED))) break;
					if (!putbin((unsigned char)(0x4a | (ss << 4)))) break;
				}
				else
					errmsg = mjrerr[EM_ILLEGAL_OPERAND];
				break;

				/* SUB r : SUB n */
			case INST_SUB:
				skip_A_comma();
				if (!arithmetic8(0x90, 0xd6)) errmsg = mjrerr[EM_ILLEGAL_OPERAND];
				break;

				/* SBC A,r : SBC A,n : SBC HL,ss */
			case INST_SBC:
				if (!look_token(&t)) break;
				if (t.type == REG_A) {
					skip_token();
					if (!get_comma()) break;
					/* A,(HL) : A,(IX+d) : A,(IY+d) */
					/* A,r : A,n */
					if (!arithmetic8(0x98, 0xde)) errmsg = mjrerr[EM_ILLEGAL_OPERAND];
				}
				else if (t.type == REG_HL) {
					int ss;
					skip_token();
					if (!get_comma()) break;
					if (!look_token(&t)) break;
					ss = chk_reg16_1(t.type, 0);
					if (ss < 0) {
						errmsg = mjrerr[EM_ILLEGAL_OPERAND];
						break;
					}
					skip_token();
					/* ED : 01 ss0 010 */
					if (!putbin((unsigned char)(0xED))) break;
					if (!putbin((unsigned char)(0x42 | (ss << 4)))) break;
				}
				else
					errmsg = mjrerr[EM_ILLEGAL_OPERAND];
				break;

				/* AND r : AND n */
			case INST_AND:
				skip_A_comma();
				if (!arithmetic8(0xa0, 0xe6)) errmsg = mjrerr[EM_ILLEGAL_OPERAND];
				break;

				/* XOR r : XOR n */
			case INST_XOR:
				skip_A_comma();
				if (!arithmetic8(0xa8, 0xee)) errmsg = mjrerr[EM_ILLEGAL_OPERAND];
				break;

				/* OR r : OR n */
			case INST_OR:
				skip_A_comma();
				if (!arithmetic8(0xb0, 0xf6)) errmsg = mjrerr[EM_ILLEGAL_OPERAND];
				break;

				/* CP r : CP n */
			case INST_CP:
				skip_A_comma();
				if (!arithmetic8(0xb8, 0xfe)) errmsg = mjrerr[EM_ILLEGAL_OPERAND];
				break;

				/* DEC r : DEC ss */
			case INST_DEC:
				opc1 = 0x0b; opc2 = 0x05;
				/* fall through */
				/* INC r : INC ss */
			case INST_INC:
				if (!opc1) { opc1 = 0x03; opc2 = 0x04; }
				if (!look_token(&t)) break;
				i = chk_reg16(t.type, &pfx);
				/* INC ss */
				if (i >= 0) {
					skip_token();
					if (pfx) if (!putbin(pfx)) break;
					if (!putbin(opc1 | (unsigned char)(i << 4))) break; /* 00 xx0 011 */
				}
				else if ((i = chk_reg8(t.type, &pfx)) >= 0) {
					skip_token();
					if (pfx) if (!putbin(pfx)) break;
					if (!putbin(opc2 | (unsigned char)(i << 3))) break;
					if (t.type == REG_IX_d_BRACED || t.type == REG_IY_d_BRACED) {
						if (!imm8()) break;
						if (!get_token(&t) || t.type != T_BRCCLOSE) break;
					}
					else if (t.type == REG_IX_BRACED || t.type == REG_IY_BRACED) {
						if (!putbin(0)) break;
					}
				}
				else
					errmsg = mjrerr[EM_ILLEGAL_OPERAND];
				break;

				/* LD ss,nn : LD (nn),HL : LD HL,(nn) : LD (nn),ss : LD ss,(nn) */
				/* LD r,r : LD r,n */
				/* LD (BC),A : LD A,(DE) : LD A,(nn) : LD (nn),A */
				/* LD A,I : LD I,A : LD R,A : LD A,R */
				/* LD SP,HL */
			case INST_LD:
				if (!look_token(&t)) break;
				if (t.type == REG_A) {
					/* A,r : A,n : A,(nn) : A,(BC|DE) : A,I : A,R */
					skip_token();
					if (!get_comma()) break;
					if (!look_token(&t)) break;
					if (t.type == REG_BC_BRACED || t.type == REG_DE_BRACED) {
						/* A,(BC) : A,(DE) */
						skip_token();
						if (!putbin(t.type == REG_BC_BRACED ? 0x0a : 0x1a)) break;
					}
					else if (t.type == REG_I || t.type == REG_R) {
						/* A,I : A,R */
						skip_token();
						if (!putbin(0xED)) break;
						if (!putbin(t.type == REG_I ? 0x57 : 0x5F)) break;
					}
					else if ((i = chk_reg8(t.type, &pfx)) >= 0) {
						/* A,r */
						skip_token();
						if (pfx) if (!putbin(pfx)) break;
						if (!putbin(0x78 | (unsigned char)i)) break;
						if (t.type == REG_IX_d_BRACED || t.type == REG_IY_d_BRACED) {
							if (!imm8()) break;
							if (!get_token(&t) || t.type != T_BRCCLOSE) break;
						}
						else if (t.type == REG_IX_BRACED || t.type == REG_IY_BRACED)
							if (!putbin(0)) break;
					}
					else if (t.type >= INST_RLC && t.type <= INST_SET) {
						/* A,RLC (IX+d) */
						skip_token();
						if (t.type <= INST_SRL) {
							if (!reg8((unsigned char)((t.type - INST_RLC) << 3) | 7, 0xCB)) break;
						}
						else {
							opc1 = (unsigned char)((t.type - INST_BIT + 1) << 6) | 7;
							goto bitinst;
						}
					}
					else if (expression()) {
						/* A,n : A,(nn) */
						if (!putbin(is_expr_braced() ? 0x3A : 0x3E)) break;
						if (is_expr_finished()) {
							nn = expr_result();
						}
						else {
							entry_unres_label(loccnt, is_expr_braced() ? PATCH_WORD : PATCH_BYTE, pinfo->linenum);
							nn = 0;
						}
						if (!putbin((unsigned char)(nn & 0xff))) break;
						if (is_expr_braced())
							if (!putbin((unsigned char)((nn >> 8) & 0xff))) break;
					}
					else
						errmsg = mjrerr[EM_ILLEGAL_OPERAND];
				}
				else if ((i = chk_reg8(t.type, &pfx)) >= 0) {
					/* r,r : r,n */
					int opr1, opcloc;
					unsigned char pfx2;
					opr1 = i;
					if (pfx) if (!putbin(pfx)) break;
					opcloc = loccnt;
					if (!putbin(0)) break; /* fill opecode later */
					skip_token();
					if (t.type == REG_IX_d_BRACED || t.type == REG_IY_d_BRACED) {
						if (!imm8()) break;
						if (!get_token(&t) || t.type != T_BRCCLOSE) break;
					}
					else if (t.type == REG_IX_BRACED || t.type == REG_IY_BRACED)
						if (!putbin(0)) break;
					if (!get_comma()) break;
					if (!look_token(&t)) break;
					if ((i = chk_reg8(t.type, &pfx2)) >= 0) {
						/* r,r */
						if (opr1 == 6 && i == 6) { /* (HL),(HL) */
							errmsg = mjrerr[EM_ILLEGAL_OPERAND];
							break;
						}
						if ((opr1 == 4 || opr1 == 5) && (i == 4 || i == 5) && pfx != pfx2) { /* IXH,IYH */
							sprintf(emstr, EJ("LD between %s%c and %s%c is not supported", "%s%c,%s%cԂLD߂݂͑܂"),
								(pfx == 0xDD ? "IX" : pfx == 0xFD ? "IY" : ""), (opr1 & 1) ? 'L' : 'H',
								(pfx2 == 0xDD ? "IX" : pfx2 == 0xFD ? "IY" : ""), (i & 1) ? 'L' : 'H');
							errmsg = emstr;
							break;
						}
						skip_token();
						if (!pfx && pfx2) { /* opr2 is index reg */
							binbuf[opcloc] = pfx2;
							opcloc = loccnt;
							if (!putbin(0)) break;
						}
						if (t.type == REG_IX_d_BRACED || t.type == REG_IY_d_BRACED) {
							if (!imm8()) break;
							if (!get_token(&t) || t.type != T_BRCCLOSE) break;
						}
						else if (t.type == REG_IX_BRACED || t.type == REG_IY_BRACED)
							if (!putbin(0)) break;
						binbuf[opcloc] = 0x40 | (unsigned char)(opr1 << 3) | (unsigned char)i;
					}
					else if (!pfx && opr1 != 6 && t.type >= INST_RLC && t.type <= INST_SET) {
						/* r,RLC (IX+d) */
						loccnt--;	/* cancel previous putbin(0) */
						locabs--;
						skip_token();
						if (t.type <= INST_SRL) {
							if (!reg8((unsigned char)((t.type - INST_RLC) << 3) | opr1, 0xCB)) break;
						}
						else {
							opc1 = (unsigned char)((t.type - INST_BIT + 1) << 6) | opr1;
							goto bitinst;
						}
					}
					else if (imm8()) {
						/* r,n */
						binbuf[opcloc] = 0x06 | (unsigned char)(opr1 << 3);
					}
					else
						errmsg = mjrerr[EM_ILLEGAL_OPERAND];
				}
				else if ((i = chk_reg16(t.type, &pfx)) >= 0) {
					/* ss,nn : ss,(nn) : SP,HL */
					skip_token();
					if (!get_comma()) break;
					if (!look_token(&t)) break;
					if (i == 3/*SP*/ && chk_reg16(t.type, &pfx) == 2/*HL*/) {
						/* SP,HL */
						skip_token();
						if (pfx) if (!putbin(pfx)) break;
						if (!putbin(0xF9)) break;
						break;
					}
					if (pfx) if (!putbin(pfx)) break;
					if (!expression()) break;
					if (is_expr_braced()) {
						if (i != 2/*HL*/) {
							if (!putbin(0xED)) break;
							if (!putbin(0x4B | (unsigned char)(i << 4))) break;
						}
						else
							if (!putbin(0x2A)) break;
					}
					else
						if (!putbin(0x01 | (unsigned char)(i << 4))) break;
					if (is_expr_finished()) {
						nn = expr_result();
					}
					else {
						entry_unres_label(loccnt, PATCH_WORD, pinfo->linenum);
						nn = 0;
					}
					if (!putbin((unsigned char)(nn & 0xff))) break;
					if (!putbin((unsigned char)((nn >> 8) & 0xff))) break;
				}
				else if (t.type == REG_BC_BRACED || t.type == REG_DE_BRACED) {
					/* (BC),A : (DE),A */
					skip_token();
					if (!get_comma()) break;
					opc1 = (t.type == REG_BC_BRACED) ? 0x02 : 0x12;
					if (!get_token(&t) || t.type != REG_A) {
						errmsg = mjrerr[EM_ILLEGAL_OPERAND];
						break;
					}
					if (!putbin(opc1)) break;
				}
				else if (t.type == REG_I || t.type == REG_R) {
					/* I,A : R,A */
					skip_token();
					if (!get_comma()) break;
					opc1 = (t.type == REG_I) ? 0x47 : 0x4F;
					if (!get_token(&t) || t.type != REG_A) {
						errmsg = mjrerr[EM_ILLEGAL_OPERAND];
						break;
					}
					if (!putbin(0xED)) break;
					if (!putbin(opc1)) break;
				}
				else if (expression() && is_expr_braced()) {
					/* (nn),A : (nn),ss */
					if (!get_comma()) break;
					if (!get_token(&t)) break;
					if (t.type == REG_A) opc1 = 0x32;
					else if ((i = chk_reg16(t.type, &pfx)) >= 0) {
						if (i == 2/*HL*/) opc1 = 0x22;
						else {
							pfx = 0xED;
							opc1 = 0x43 | (unsigned char)(i << 4);
						}
					}
					else {
						errmsg = mjrerr[EM_ILLEGAL_OPERAND];
						break;
					}
					if (pfx) if (!putbin(pfx)) break;
					if (!putbin(opc1)) break;
					if (is_expr_finished()) {
						nn = expr_result();
					}
					else {
						entry_unres_label(loccnt, is_expr_braced() ? PATCH_WORD : PATCH_BYTE, pinfo->linenum);
						nn = 0;
					}
					if (!putbin((unsigned char)(nn & 0xff))) break;
					if (!putbin((unsigned char)((nn >> 8) & 0xff))) break;
				}
				else
					errmsg = mjrerr[EM_ILLEGAL_OPERAND];
				break;

				/* JR e, JR cond,e */
			case INST_JR:
				if (!look_token(&t)) break;
				i = chk_cond4(t.type);
				if (i >= 0) { /* conditional */
					opc1 = 0x20 | (i << 3);
					skip_token();
					if (!get_comma()) break;
				}
				else opc1 = 0x18; /* non-conditional */
				if (!putbin(opc1)) break;
				if (!relative_address()) break;
				break;

				/* DJNZ e */
			case INST_DJNZ:
				if (!putbin(0x10)) break;
				if (!relative_address()) break;
				break;

				/* JP nn : JP cond,nn : JP (HL) */
			case INST_JP:
				if (!look_token(&t)) break;
				if (t.type == REG_HL_BRACED) {/*nothing*/ }
				else if (t.type == REG_IX_BRACED) pfx = 0xDD;
				else if (t.type == REG_IY_BRACED) pfx = 0xFD;
				else {
					opc1 = 0xC2;
					opc2 = 0xC3;
					goto jpcallinst;
				}
				skip_token();
				if (pfx) if (!putbin(pfx)) break;
				if (!putbin(0xE9)) break;
				break;

				/* CALL nn : CALL cond,nn */
			case INST_CALL:
				opc1 = 0xC4;
				opc2 = 0xCD;
				if (!look_token(&t)) break;
			jpcallinst:		i = chk_cond8(t.type);
				if (i >= 0) { /* conditional */
					opc1 |= (unsigned char)(i << 3);
					skip_token();
					if (!get_comma()) break;
				}
				else opc1 = opc2; /* non-conditional */
				if (!putbin(opc1)) break;
				if (!imm16()) break;
				break;

				/* RET : RET cond */
			case INST_RET:
				if (!look_token(&t)) break;
				i = chk_cond8(t.type);
				if (i >= 0) { /* conditional */
					opc1 = 0xC0 | (unsigned char)(i << 3);
					skip_token();
				}
				else opc1 = 0xC9; /* non-conditional */
				if (!putbin(opc1)) break;
				break;

				/* RETN: RETI */
			case INST_RETN:
				if (!putbin(0xED)) break;
				if (!putbin(0x45)) break;
				break;
			case INST_RETI:
				if (!putbin(0xED)) break;
				if (!putbin(0x4D)) break;
				break;

				/* PUSH/POP {BC|DE|HL|AF|IX|IY} */
			case INST_PUSH:
				opc1 = 0xC5;
				goto pushpopinst;
			case INST_POP:
				opc1 = 0xC1;
			pushpopinst:		if (!get_token(&t)) break;
				if ((i = chk_reg16_2(t.type, &pfx)) >= 0) {
					if (pfx) if (!putbin(pfx)) break;
					if (!putbin(opc1 | (unsigned char)(i << 4))) break;
				}
				else
					errmsg = mjrerr[EM_ILLEGAL_OPERAND];
				break;

				/* NOP */
			case INST_NOP:
				if (!putbin(0x00)) break;
				break;

				/* DAA */
			case INST_DAA:
				if (!putbin(0x27)) break;
				break;

				/* CPL */
			case INST_CPL:
				if (!putbin(0x2F)) break;
				break;

				/* NEG */
			case INST_NEG:
				if (!putbin(0xED)) break;
				if (!putbin(0x44)) break;
				break;

				/* SCF : CCF */
			case INST_SCF:
				if (!putbin(0x37)) break;
				break;
			case INST_CCF:
				if (!putbin(0x3F)) break;
				break;

				/* HALT */
			case INST_HALT:
				if (!putbin(0x76)) break;
				break;

				/* EXX */
			case INST_EXX:
				if (!putbin(0xD9)) break;
				break;

				/* EX AF,AF' : EX DE,HL : EX (SP),HL */
			case INST_EX:
				if (!get_token(&t)) break;
				if (t.type == REG_AF) {
					/* AF,AF' */
					if (!get_comma()) break;
					if (!get_token(&t) || (t.type != REG_AF_ && t.type != REG_AF)) {
						errmsg = mjrerr[EM_ILLEGAL_OPERAND];
						break;
					}
					if (!putbin(0x08)) break;
				}
				else if (t.type == REG_DE) {
					/* DE,HL */
					if (!get_comma()) break;
					if (!get_token(&t) || t.type != REG_HL) {
						errmsg = mjrerr[EM_ILLEGAL_OPERAND];
						break;
					}
					if (!putbin(0xEB)) break;
				}
				else if (t.type == REG_SP_BRACED) {
					/* (SP),HL : (SP),IX : (SP),IY */
					if (!get_comma()) break;
					if (!get_token(&t) || chk_reg16(t.type, &pfx) != 2) {
						errmsg = mjrerr[EM_ILLEGAL_OPERAND];
						break;
					}
					if (pfx) if (!putbin(pfx)) break;
					if (!putbin(0xE3)) break;
				}
				else {
					errmsg = mjrerr[EM_ILLEGAL_OPERAND];
					break;
				}
				break;

				/* RLCA : RRCA : RLA : RRA */
			case INST_RLCA:
				if (!putbin(0x07)) break;
				break;
			case INST_RRCA:
				if (!putbin(0x0F)) break;
				break;
			case INST_RLA:
				if (!putbin(0x17)) break;
				break;
			case INST_RRA:
				if (!putbin(0x1F)) break;
				break;

				/* RLC : RRC : RL : RR : SLA : SRA : SLL : SRL */
			case INST_RLC:
			case INST_RRC:
			case INST_RL:
			case INST_RR:
			case INST_SLA:
			case INST_SRA:
			case INST_SLL:
			case INST_SRL:
				reg8((unsigned char)((t.type - INST_RLC) << 3) | 6, 0xCB);
				break;

				/* BIT : RES : SET */
			case INST_SET:
			case INST_RES:
			case INST_BIT:
				opc1 = (unsigned char)((t.type - INST_BIT + 1) << 6) | 6;
			bitinst:		if (!expression()) break;
				if (!get_comma()) break;
				if (is_expr_finished()) {
					nn = expr_result();
					if (nn < 0 || nn > 7) {
						errmsg = mjrerr[EM_BITCNT_OVERFLOW];
						break;
					}
					opc1 |= (unsigned char)(nn << 3);
				}
				reg8(opc1, 0xCB);
				if (!is_expr_finished())
					entry_unres_label(loccnt - 1, PATCH_BITCNT, pinfo->linenum);
				break;

				/* LDI : LDD : LDIR : LDDR */
			case INST_LDI:
				if (!putbin(0xED)) break;
				if (!putbin(0xA0)) break;
				break;
			case INST_LDD:
				if (!putbin(0xED)) break;
				if (!putbin(0xA8)) break;
				break;
			case INST_LDIR:
				if (!putbin(0xED)) break;
				if (!putbin(0xB0)) break;
				break;
			case INST_LDDR:
				if (!putbin(0xED)) break;
				if (!putbin(0xB8)) break;
				break;

				/* CPI : CPD : CPIR : CPDR */
			case INST_CPI:
				if (!putbin(0xED)) break;
				if (!putbin(0xA1)) break;
				break;
			case INST_CPD:
				if (!putbin(0xED)) break;
				if (!putbin(0xA9)) break;
				break;
			case INST_CPIR:
				if (!putbin(0xED)) break;
				if (!putbin(0xB1)) break;
				break;
			case INST_CPDR:
				if (!putbin(0xED)) break;
				if (!putbin(0xB9)) break;
				break;

				/* IN A,(n) : IN r,(C) */
			case INST_IN:
				if (!get_token(&t)) break;
				if (((i = chk_reg8(t.type, &pfx)) >= 0 && !pfx) ||
					(t.type == IDENTIFIER && t.tstrlen == 1 &&
						(t.tstr[0] == 'F' || t.tstr[0] == 'f'))) {
					if (t.type == IDENTIFIER) i = 6; /* IN F,(C) */
					if (!get_comma()) break;
					if (!look_token(&t)) {
						errmsg = mjrerr[EM_ILLEGAL_OPERAND];
						break;
					}
					if (t.type == REG_C_BRACED) {
						/* IN r,(C) */
						if (!putbin(0xED)) break;
						if (!putbin(0x40 | (unsigned char)(i << 3))) break;
						skip_token();
					}
					else {
						/* IN A,(n) */
						if (i != 7/*REG_A*/) {
							errmsg = mjrerr[EM_ILLEGAL_OPERAND];
							break;
						}
						if (!putbin(0xDB)) break;
						if (!imm8()) break;
						if (!is_expr_braced()) {
							errmsg = mjrerr[EM_ILLEGAL_OPERAND];
							break;
						}
					}
				}
				else
					errmsg = mjrerr[EM_ILLEGAL_OPERAND];
				break;

				/* TSTI (C) */
			case INST_TSTI:
				if (!get_token(&t)) break;
				if (t.type == REG_C_BRACED) {
					/* TSTI (C) */
					if (!putbin(0xED)) break;
					if (!putbin(0x70)) break;
				}
				else
					errmsg = mjrerr[EM_ILLEGAL_OPERAND];
				break;

				/* OUT (n),A : OUT (C),r */
			case INST_OUT:
				if (!look_token(&t)) break;
				if (t.type == REG_C_BRACED) {
					/* OUT (C),r */
					skip_token();
					if (!get_comma()) break;
					if (!get_token(&t)) break;
					if ((i = chk_reg8(t.type, &pfx)) >= 0 && !pfx && i != 6/*(HL)*/) {
						if (!putbin(0xED)) break;
						if (!putbin(0x41 | (unsigned char)(i << 3))) break;
					}
					else if (t.type == IMMIDIATE && t.value == 0) {
						if (!putbin(0xED)) break;
						if (!putbin(0x71)) break;
					}
					else
						errmsg = mjrerr[EM_ILLEGAL_OPERAND];
				}
				else if (expression() && is_expr_braced()) {
					/* OUT (n),A */
					if (!get_comma()) break;
					if (!get_token(&t) || t.type != REG_A) {
						errmsg = mjrerr[EM_ILLEGAL_OPERAND];
						break;
					}
					if (!putbin(0xD3)) break;
					if (is_expr_finished()) {
						nn = expr_result();
					}
					else {
						entry_unres_label(loccnt, PATCH_BYTE, pinfo->linenum);
						nn = 0;
					}
					if (!putbin((unsigned char)(nn & 0xff))) break;
				}
				else
					errmsg = mjrerr[EM_ILLEGAL_OPERAND];
				break;

				/* INI : IND : INIR : INDR */
			case INST_INI:
				if (!putbin(0xED)) break;
				if (!putbin(0xA2)) break;
				break;
			case INST_IND:
				if (!putbin(0xED)) break;
				if (!putbin(0xAA)) break;
				break;
			case INST_INIR:
				if (!putbin(0xED)) break;
				if (!putbin(0xB2)) break;
				break;
			case INST_INDR:
				if (!putbin(0xED)) break;
				if (!putbin(0xBA)) break;
				break;

				/* OUTI : OUTD : OTIR : OTDR */
			case INST_OUTI:
				if (!putbin(0xED)) break;
				if (!putbin(0xA3)) break;
				break;
			case INST_OUTD:
				if (!putbin(0xED)) break;
				if (!putbin(0xAB)) break;
				break;
			case INST_OTIR:
				if (!putbin(0xED)) break;
				if (!putbin(0xB3)) break;
				break;
			case INST_OTDR:
				if (!putbin(0xED)) break;
				if (!putbin(0xBB)) break;
				break;

				/* DI : EI */
			case INST_DI:
				if (!putbin(0xF3)) break;
				break;
			case INST_EI:
				if (!putbin(0xFB)) break;
				break;

				/* RST n */
			case INST_RST:
				if (!expression()) break;
				if (!is_expr_finished()) {
					errmsg = mjrerr[EM_UNRESOLVED_EXPR];
					break;
				}
				if (expr_result() & 0xFFFFFFC7) {
					errmsg = mjrerr[EM_ILLEGAL_OPERAND];
					break;
				}
				if (!putbin(0xC7 | (unsigned char)(expr_result() & 0x38))) break;
				break;

				/* RRD : RLD */
			case INST_RRD:
				if (!putbin(0xED)) break;
				if (!putbin(0x67)) break;
				break;
			case INST_RLD:
				if (!putbin(0xED)) break;
				if (!putbin(0x6F)) break;
				break;

				/* IM n */
			case INST_IM:
				if (!expression()) break;
				if (!is_expr_finished()) {
					errmsg = mjrerr[EM_UNRESOLVED_EXPR];
					break;
				}
				if (!putbin(0xED)) break;
				switch (expr_result()) {
				case 0:
					putbin(0x46);
					break;
				case 1:
					putbin(0x56);
					break;
				case 2:
					putbin(0x5E);
					break;
				default:
					errmsg = mjrerr[EM_ILLEGAL_OPERAND];
					break;
				}
				break;

			default:
				errmsg = EJ("Unsupported instruction", "T|[g̖");
				break;
			}
			locdol = locabs;
		}
		else if (CMD_START <= t.type && t.type <= CMD_END) {
			switch (t.type) {
				unsigned char c;
				/* ORG */
			case CMD_ORG:
				if (!expression()) break;
				if (!is_expr_finished()) {
					errmsg = mjrerr[EM_UNRESOLVED_EXPR];
					break;
				}
				if (!pi->pass2 && using_in_page_used && !fix_in_page_address()) break;
				locabs = expr_result();
				locabs_start = locabs;
				loccnt_start = loccnt;
				if (!load_addr) load_addr = locabs;
				if (!exec_addr) exec_addr = locabs;
				//			if (using_in_page_used) {
				//			    errmsg = EJ("Multiple ORG cannot be used after IN_PAGE/PAGE_ALIGNED",
				//				        "ORG[߂IN_PAGE/PAGE_ALIGNED̕p͂ł܂");
				//			}
				break;
			case CMD_END:
				if (supress_putbin) {
					/* finish OFFSET definition */
					supress_putbin = 0;
					locabs = sav_locabs;
					break;
				}
				if (arr_num) {
					/* finish ARRANGE operation */
					arr_num = 0;
					locabs = sav_locabs;
					loccnt = sav_loccnt;
					loccnt_s = loccnt;
					locabs_s = locabs;
					copyto = t.from;
					break;
				}
				if (!look_token(&t) || t.type == 0) return PRES_END;
				if (expression()) {
					if (is_expr_finished()) {
						exec_addr = expr_result();
						return PRES_END;
					}
					else
						errmsg = mjrerr[EM_UNRESOLVED_EXPR];
				}
				break;
			case CMD_OFFSET:
				if (!supress_putbin) {
					if (!check_special_putbin("OFFSET")) break;
					supress_putbin = 1;
					sav_locabs = locabs;
				}
				if (!look_token(&t) || t.type == 0) {
					locabs = 0;
				}
				else {
					if (!expression()) break;
					if (!is_expr_finished()) {
						errmsg = mjrerr[EM_UNRESOLVED_EXPR];
						break;
					}
					locabs = expr_result() & 0x0000FFFF;
				}
				break;
			case CMD_DB:
				do {
					if (!look_token(&t)) break;
					if (t.type == T_DQUOTE) {
						int loccnt_last;
						skip_token();
						while ((c = read1chr()) != 0x22 && c != 0) {
							loccnt_last = loccnt;
							if (!putbin_transcode(c)) break;
						}
						if (errmsg) break;
						if (db_encode == DB_ASCII_MZ_JP && small_hira) {
							loccnt_last = loccnt;
							if (!putbin(0x05)) break;
							small_hira = 0;
						}
						if (!c) {
							errmsg = mjrerr[EM_MISSING_CLOSE_QUOTE];
							break;
						}
						if (look_token(&t) && t.type == OPR_PLUS && skip_token()) {
							if (!expression()) break;
							if (!is_expr_finished()) {
								errmsg = mjrerr[EM_UNRESOLVED_EXPR];
								break;
							}
							if (!pi->pass2) binbuf[loccnt_last] += (unsigned char)(expr_result() & 0x000000FF);
						}
					}
					else {
						if (!imm8()) break;
					}
				} while (look_token(&t) && t.type == T_COMMA && skip_token());
				break;
			case CMD_DW:
				do {
					if (!imm16()) break;
				} while (look_token(&t) && t.type == T_COMMA && skip_token());
				break;
			case CMD_DS:
				if (!expression()) break;
				if (!is_expr_finished()) {
					errmsg = mjrerr[EM_UNRESOLVED_EXPR];
					break;
				}
				i = expr_result();
				if (i < 0) {
					errmsg = EJ("Negative size is specified on DS", "DSɕ̃TCY͎wł܂");
					break;
				}
				c = 0;
				if (look_token(&t) && t.type == T_COMMA) {
					skip_token();
					if (!expression()) break;
					if (!is_expr_finished()) {
						errmsg = mjrerr[EM_UNRESOLVED_EXPR];
						break;
					}
					c = (unsigned char)(expr_result() & 0x00ff);
				}
				for (; i; i--)
					if (!putbin(c)) break;
				break;
			case CMD_ALIGN:
				if (expression() && is_expr_finished()) {
					int j, k, l;
					i = expr_result();
					if (i > 256) {
						errmsg = EJ("alignment value too big", "ALIGN̎wl傫܂");
						break;
					}
					for (j = 0, k = 0, l = 1; j < 9; j++) {
						if (i & l) k++;
						l <<= 1;
					}
					if (k != 1) {
						errmsg = EJ("Alignment value should be 2^n", "ALIGN̎wl2^nɂĂ");
						break;
					}
					i--;
					while (locabs & i)
						if (!putbin(0)) break;
				}
				else
					errmsg = mjrerr[EM_UNRESOLVED_EXPR];
				break;
			case CMD_BINCLUDE:
			case CMD_INCLUDE: {
				char* p;
				if (pi->pass2) {
					if (doflush == 0 && outasm[0]) {
						doflush = 1;
						copyto = rp = t.from; /* rewind */
						goto flush_list;
					}
					else {
						doflush = 0;
					}
				}
				skipspace();
				p = pinfo->buf;
				errmsg = EJ("Too long filename", "t@C܂");
				if (peek1chr() == '\"') {
					read1chr();
					for (i = 0; i < 256; i++) {
						c = read1chr();
						if (!c) {
							errmsg = mjrerr[EM_MISSING_CLOSE_DC];
							break;
						}
						if (c == '\"') c = 0;
						*p++ = c;
						if (!c) {
							token t2;
						incl_check:				if (!get_token(&t2) || t2.type != 0) {
							sprintf(emstr, mjrerr[EM_SOMETHING_TAILED],
								t.type == CMD_INCLUDE ? "INCLUDE" : "BINCLUDE");
							errmsg = emstr;
							break;
						}
						return (t.type == CMD_INCLUDE) ? PRES_INCLUDE : PRES_BINCLUDE;
						}
					}
				}
				else {
					for (i = 0; i < 256; i++) {
						c = read1chr();
						if (isspace(c) || c == ';') c = 0;
						*p++ = c;
						if (!c) goto incl_check;
					}
				}
				break;
			}
			case CMD_ERROR: {
				char* p;
				skipspace();
				if (read1chr() != '\"') {
					errmsg = mjrerr[EM_MISSING_OPEN_DC];
					break;
				}
				for (i = MAXTOKENLEN, p = t.tstr; i; i--) {
					c = read1chr();
					if (!c || c == '\"') break;
					*p++ = c;
				}
				if (!i) {
					errmsg = EJ("Too long error message", "G[bZ[W܂");
					break;
				}
				if (c != '\"') {
					errmsg = mjrerr[EM_MISSING_CLOSE_DC];
					break;
				}
				*p = 0;
				errmsg = t.tstr;
				break;
			}
			case CMD_ENDM: /* just ignore it... */
				copyto = t.from;
				break;
			case CMD_IF:
			proc_if:		if (!expression()) break;
				if (is_expr_finished()) {
					i = (expr_result() != 0) ? IFSTT_SEARCH_ELSE : IFSTT_SKIP_UNTIL_ELSE;
				}
				else {
					errmsg = mjrerr[EM_UNRESOLVED_EXPR];
					i = IFSTT_SKIP_ALL; /* skip inside if the condition is invalid */
				}
				if ((pinfo->ifsp + 1) == MAX_IFSTK) {
					errmsg = mjrerr[EM_IF_TOO_NESTED];
					i = IFSTT_SKIP_ALL;
				}
				pinfo->ifstk[(pinfo->ifsp)++] = pinfo->ifstt;
				pinfo->ifstt = i;
				copyto = t.from;
				break;
			case CMD_ELSE:
				switch (pinfo->ifstt) {
				case IFSTT_SEARCH_ELSE:
					/* IF <true> ... ELSE */
					pinfo->ifstt = IFSTT_SKIP_UNTIL_ENDIF;
					break;
				case IFSTT_SKIP_UNTIL_ELSE:
					/* IF <false> ... ELSE */
					pinfo->ifstt = IFSTT_JUST_IGNORE_ENDIF;
					break;
				case IFSTT_SKIP_ALL:
					break;
				default:
					errmsg = EJ("ELSE without IF", "ELSȆOIF܂");
					break;
				}
				copyto = t.from;
				break;
			case CMD_ELIF:
				switch (pinfo->ifstt) {
				case IFSTT_SEARCH_ELSE:
					/* IF <true> ... ELIF */
					pinfo->ifstt = IFSTT_SKIP_UNTIL_ENDIF;
					break;
				case IFSTT_SKIP_UNTIL_ELSE:
					/* IF <false> ... ELIF */
					pinfo->ifstt = pinfo->ifstk[--(pinfo->ifsp)];
					goto proc_if;
				case IFSTT_SKIP_UNTIL_ENDIF:
				case IFSTT_SKIP_ALL:
					break;
				default:
					errmsg = EJ("ELIF without IF", "ELIF̑OIF܂");
					break;
				}
				copyto = t.from;
				break;
			case CMD_ENDIF:
				if (pinfo->ifstt != IFSTT_NONE)
					pinfo->ifstt = pinfo->ifstk[--(pinfo->ifsp)];
				else
					errmsg = EJ("ENDIF without IF", "ENDIF̑OIF܂");
				copyto = t.from;
				break;
			case CMD_DISPC:
				do {
					if (!look_token(&t)) break;
					if (t.type == T_SQUOTE || t.type == T_DQUOTE) {
						char q = (t.type == T_SQUOTE) ? 0x27 : 0x22;
						skip_token();
						while ((c = read1chr()) != q && c != 0) {
							i = (int)(c & 0x000000ff);
							if (i >= 0x20 && i <= 0x7f) c = dispc[i - 0x20];
							else if (i >= 0xa0 && i <= 0xdf) c = dispck[i - 0xa0];
							if (!putbin(c)) break;
						}
						if (!c) {
							errmsg = mjrerr[EM_MISSING_CLOSE_QUOTE];
							break;
						}
					}
					else {
						if (!imm8()) break;
					}
				} while (look_token(&t) && t.type == T_COMMA && skip_token());
				break;
			case CMD_PROC:
				if (pi->pass2) {
					dig_namespace_pass2();
					skip_token();
					/*copyfrom = rp;*/
					break;
				}
				if (!dig_namespace(0)) {
					break;
				}
				break;
			case CMD_ENDP:
				/*if (pi->pass2) copyfrom = rp;*/
				if (using_in_page_type) finish_using_in_page();
				if (!leave_namespace()) {
					errmsg = EJ("Tried to exit the base level procedure block", "ENDP̑OPROC܂");
					break;
				}
				break;
			case CMD_CHRMAP: {
				char* p;
				skipspace();
				if (read1chr() != '\"') {
					errmsg = mjrerr[EM_MISSING_OPEN_DC];
					break;
				}
				for (i = MAXTOKENLEN, p = t.tstr; i; i--) {
					c = read1chr();
					if (!c || c == '\"') break;
					*p++ = c;
				}
				*p = 0;
				if (c != '\"') {
					errmsg = mjrerr[EM_MISSING_CLOSE_DC];
					break;
				}
				if (!strcmpi(t.tstr, "ASCII")) {
					db_encode = DB_ASCII_GENERIC;
					break;
				}
				else if (!strcmpi(t.tstr, "MZ-ASCII") ||
					!strcmpi(t.tstr, "MZ-ASCII-JP")) {
					db_encode = DB_ASCII_MZ_JP;
					break;
				}
				else if (!strcmpi(t.tstr, "MZ-ASCII-EU")) {
					db_encode = DB_ASCII_MZ_EU;
					break;
				}
				else if (!strcmpi(t.tstr, "DISPLAYCODE") ||
					!strcmpi(t.tstr, "DISPLAYCODE-JP")) {
					db_encode = DB_DISP_JP;
					break;
				}
				else if (!strcmpi(t.tstr, "DISPLAYCODE-EU")) {
					db_encode = DB_DISP_EU;
					break;
				}
				else if (!strcmpi(t.tstr, "USER-DEFINED")) {
					db_encode = DB_USER_DEFINED;
					break;
				}
				else {
					errmsg = EJ("Unsupported encode directive", "m̃LN^}bvw蕶");
				}
				break;
			}
			case CMD_MAPCHR: {  // MAPCHR <ASCII code of destination>, <value to assign> [,<value to assign> ...]. /
								// MAPCHR <蓖ĐASCIIR[h>, <蓖Ăl> [,<蓖Ăl> c]
				int chrcode;
				skipspace();
				if (!expression()) break;
				if (!is_expr_finished()) {
					errmsg = mjrerr[EM_UNRESOLVED_EXPR];
					break;
				}
				chrcode = expr_result() & 0x000000FF;
				while (look_token(&t) && t.type == T_COMMA && skip_token()) {
					if (!expression()) break;
					if (!is_expr_finished()) {
						errmsg = mjrerr[EM_UNRESOLVED_EXPR];
						break;
					}
					mapchr[chrcode++] = expr_result() & 0x000000FF;
					chrcode &= 0x000000FF;
				}
				break;
			}
			case CMD_ARRANGE: {
				int loctmp;
				if (!check_special_putbin("ARRANGE")) break;
				/* Placement start address / zuJnAhX */
				if (!expression()) break;
				if (!is_expr_finished()) {
					errmsg = mjrerr[EM_UNRESOLVED_EXPR];
					break;
				}
				loctmp = expr_result() & 0x0000FFFF;
				if (loctmp < locabs_start || loctmp >= locabs_start + loccnt - loccnt_start) {
					errmsg = EJ("ARRANGE out of range", "ARRANGEzuwAhX͈͊O");
					break;
				}
				arr_index = arr_num = arr_count = 0;
				while (arr_num < 64 && look_token(&t) && t.type == T_COMMA && skip_token()) {
					/* Specify the number of bytes to be placed / zuoCgw */
					if (!expression()) break;
					if (!is_expr_finished()) {
						errmsg = mjrerr[EM_UNRESOLVED_EXPR];
						break;
					}
					arr_bytes[arr_num++] = expr_result() & 0x0000FFFF;
				}
				if (arr_num & 1) {
					errmsg = EJ("ARRANGE data pair broken", "ARRANGEzuoCgwgɂȂĂȂ");
					break;
				}
				sav_locabs = locabs;
				sav_loccnt = loccnt;
				locabs = loctmp;
				loccnt = (loctmp - locabs_start) + loccnt_start;
				break;
			}
			case CMD_REPT: {		// REPT {Formula / } (macro definition) ENDM
				unsigned long li;
				label* l;
				char tmp[32];
				token t2;
				if (pi->rept_cnt) {
					errmsg = EJ("Nested REPT", "REPT̒`REPT͒`ł܂");
					break;
				}
				sprintf(tmp, "?R?00000");
				long2cstr(&tmp[3], rept_num++);
				/* Definition / ` */
				if (!(li = getlabel(tmp))) {
					if (!(li = entrylabel(tmp))) break;
					l = getlabelptr(li);
					l->defstate = LABEL_MACRO;
				}
				strcpy(pi->buf, tmp);
				/* Number of repetitions / JԂ */
				if (!expression()) break;
				if (!is_expr_finished()) {
					errmsg = mjrerr[EM_UNRESOLVED_EXPR];
					break;
				}
				pi->rept_cnt = expr_result();
				if (!get_token(&t2) || t2.type != 0) {
					sprintf(emstr, mjrerr[EM_SOMETHING_TAILED], "REPT");
					errmsg = emstr;
					break;
				}
				return PRES_MACRO;
			}
			default:
				errmsg = EJ("Unsupported assembler command", "T|[g̋[");
				break;
			}
		}
		else if (t.type == IDENTIFIER) {
			token t2;
			unsigned long li;
			label* l;
			if (t.value == IDENTIFIER_REF_ONLY) {
				errmsg = EJ("Illegal label identifier", "ȃxʎq");
				break;
			}
			/* macro? */
			li = getlabel(t.tstr);
			if (li) {
				l = getlabelptr(li);
			ext_macro:
				if (l->defstate == LABEL_MACRO) {
					parseinfo pinfo_m;
					if (pi->pass2) {
						if (!look_token(&t2)) break;
						if (t2.type == CMD_MACRO) return PRES_MACRO;
						if (doflush == 0 && outasm[0]) {
							doflush = 1;
							copyto = rp = t.from; /* rewind */
							goto flush_list;
						}
						else {
							doflush = 0;
						}
					}
					if (macro_nest_lev >= MAXMACRONEST) {
						errmsg = EJ("Macro nested too much", "}ÑlXg[܂");
						break;
					}
					macro_nest_lev++;
					li = l->value;
					pinfo_m.linenum = 1;
					pinfo_m.filename = pi->filename;
					pinfo_m.srcfile = NULL;
					pinfo_m.ifstt = IFSTT_NONE;
					pinfo_m.ifsp = 0;
					pinfo_m.pass2 = pi->pass2;
					pinfo_m.rept_cnt = 0;
					pinfo_m.srclinebuf = l->value;
					get_macro_args(&pinfo_m);
					if (parse_text(&pinfo_m) != PRES_OK) {
						errmsg = EJ("Error in macro", "}N̓WJɃG[܂");
					}
					if (macro_nest_lev) macro_nest_lev--;
					skip_until_eol();
					copyfrom = rp;
					break;
				}
			}
			if (!look_token(&t2)) break;
			switch (t2.type) {
			case CMD_LABELDEF_GLOBAL:	// label::	(global label definition)
				if (pi->pass2) {
					i = get_namespace_id();
					topmost_namespace();
					li = getlabel(t.tstr);
					l = getlabelptr(li);
					l->reftype |= LF_APPEARED_IN_PASS2;
					switch_namespace(i);
					skip_token();
					break;
				}
				i = get_namespace_id();
				topmost_namespace();
				if (!(li = define_label(t.tstr, 1))) break;
				skip_token();
				define_location_label(li);
				l = getlabelptr(li);
				l->reftype |= LF_GLOBAL;
				switch_namespace(i);
				break;
			case CMD_LABELDEF:		// label:	(normal label definition)
				if (pi->pass2) {
					li = getlabel(t.tstr);
					l = getlabelptr(li);
					l->reftype |= LF_APPEARED_IN_PASS2;
					skip_token();
					break;
				}
				if (!(li = define_label(t.tstr, 0))) break;
				skip_token();
				define_location_label(li);
				/*dbg printf("labeldef %s = %04X\n", t.tstr,l->value);*/
				break;
			case CMD_R_EQU:		// label = x	(re-definable label)
				if (pi->pass2) {
					li = getlabel(t.tstr);
					l = getlabelptr(li);
					l->reftype |= LF_APPEARED_IN_PASS2;
					labelval_pass2 = (int)(l->value) | 0x00010000;
					skip_token();
					expression();
					break;
				}
				if (!(li = getlabel(t.tstr))) {
					if (!(li = entrylabel(t.tstr))) break;
				}
				else {
					l = getlabelptr(li);
					if (!(l->reftype & LF_REDEFINABLE)) {
						errmsg = EJ("Label is not redefinable", "x͍Ē`s");
						break;
					}
				}
				skip_token();
				if (!expression()) break;
				if (!is_expr_finished()) {
					errmsg = mjrerr[EM_UNRESOLVED_EXPR];
					break;
				}
				l = getlabelptr(li);
				l->value = (unsigned short)expr_result();
				l->defstate = LABEL_DEFINED;
				l->reftype |= LF_REDEFINABLE;
				break;
			case CMD_EQU:		// label EQU x	(normal label definition)
				if (pi->pass2) {
					li = getlabel(t.tstr);
					l = getlabelptr(li);
					l->reftype |= LF_APPEARED_IN_PASS2;
					labelval_pass2 = (int)(l->value) | 0x00010000;
					skip_token();
					expression();
					break;
				}
				if (!(li = define_label(t.tstr, 0))) break;
				skip_token();
				if (!expression()) break;
				if (is_expr_finished()) {
					l = getlabelptr(li);
					l->value = (unsigned short)expr_result();
					l->defstate = LABEL_DEFINED;
				}
				else {
					entry_unres_label(li, PATCH_LABEL, pinfo->linenum);
				}
				break;
			case CMD_MACRO:		// label MACRO	(macro definition)
				if (!(li = getlabel(t.tstr))) {
					if (!(li = entrylabel(t.tstr))) break;
				}
				else {
					l = getlabelptr(li);
					if (l->defstate != LABEL_UNDEFINED) {
						em_dup_label(t.tstr, l->defstate);
						break;
					}
				}
				l = getlabelptr(li);
				l->defstate = LABEL_MACRO;
				l->value = 0;
				strcpy(pinfo->buf, t.tstr);
				skip_token();
				if (!get_token(&t2) || t2.type != 0) {
					sprintf(emstr, mjrerr[EM_SOMETHING_TAILED], "MACRO");
					errmsg = emstr;
					break;
				}
				return PRES_MACRO;
			case CMD_PROC:		// label PROC	(dig namespace)
				if (pi->pass2) {
					dig_namespace_pass2();
					skip_token();
					/* replace "label PROC" to "label:" */
	//			    strncpy(wp, t.tstr, t.tstrlen);
	//			    wp += t.tstrlen;
	//			    *wp++ = ':';
	//			    copyfrom = rp;
					break;
				}
				if (!(li = define_label(t.tstr, 0))) break;
				skip_token();
				l = getlabelptr(li);
				l->value = locabs;
				l->defstate = LABEL_DEFINED;
				if (!dig_namespace(t.tstr)) {
					break;
				}
				break;
			case CMD_IN_PAGE:        // label IN_PAGE      { ... }
			case CMD_PAGE_ALIGNED: { // label PAGE_ALIGNED { ... }
				label* l;
				char reftype = 0;
				int r;
				while (t2.type == CMD_IN_PAGE || t2.type == CMD_PAGE_ALIGNED) {
					if (t2.type == CMD_IN_PAGE) {
						if (!(r = check_special_putbin("IN_PAGE"))) break;
						using_in_page_upto256 = 1;
						if (reftype == 0) reftype = LF_IN_PAGE;
					}
					else {
						if (!(r = check_special_putbin("PAGE_ALIGNED"))) break;
						reftype = LF_PAGE_ALIGNED;
					}
					skip_token();
					r = look_token(&t2);
				}
				if (!r) break;
				if (pi->pass2) {
					l = getlabelptr(getlabel(t.tstr));
					sav_locabs = locabs;
					sav_loccnt = loccnt;
					loccnt_s = loccnt = (l->value - locabs) + loccnt;
					locabs_s = locabs = l->value;
					locdol = locabs;
					using_in_page_upto256 = 0; // no check //
				}
				else {
					if (!(li = define_label(t.tstr, 1))) break;
					l = getlabelptr(li);
					sav_locabs = locabs;
					sav_loccnt = loccnt;
					using_in_page = li;
					locabs = 0;
					loccnt = loccnt_in_page;
					locdol = 0;
					l->value = loccnt_in_page - MAXBINBUF;
					l->defstate = LABEL_UNDEFINED;
					l->reftype = reftype;
				}
				using_in_page_type = 1; /* until EOL */
				using_in_page_used = 1;
				if (r && t2.type == CMD_PROC) {
					skip_token();
					using_in_page_type = 2; /* until ENDP */
					if (pi->pass2) {
						dig_namespace_pass2();
						break;
					}
					else {
						if (!dig_namespace(t.tstr)) break;
					}
				}
				break;
			}
			default:
				/* undefied label without any definition directive */
				if (get_namespace_id()) {
					/* if we are in proc-block, macro may be defined outside */
					l = getlabelptr_outside(t.tstr, LABEL_MACRO);
					if (l) goto ext_macro;	// macro of this name is defined outside
				}
				sprintf(emstr, EJ("Undefined identifier '%s'", "`̎ʎq '%s'"), t.tstr);
				errmsg = emstr;
				break;
			}
		}
		else if (t.type == DELTEXT) {
			/* deleted text */
			copyfrom = rp;
			t.type = 0;
		}
		else if (t.type != 0/*EOS*/)
			errmsg = mjrerr[EM_SYNTAX_ERROR];
	flush_list:
		if (pi->pass2) {
			if (!copyto) {
				copyto = rp;
			}
			else {
				didcut = 1;
			}
			i = copyto - copyfrom;
			strncpy(wp, copyfrom, i);
			wp += i;
			*wp = 0;
			copyfrom = rp;
			if (doflush) break;
		}
		if (t.type == 0) {
			// reached to EOL
			loccnt_e = loccnt;
			if (using_in_page_type == 1) finish_using_in_page();
			break;
		}
	}
	if (pinfo->pass2 && !arr_num) {
		static int blankline = 0;
		int allspace;
		char tmpbuf[256], * p, c, * p0;
		for (p = p0 = outasm, allspace = 1; (c = *p) != 0; p++) {
			if (!isspace(c)) {
				p0 = p; /* last non-space char */
				allspace = 0;
			}
		}
		if (p0 < p) p0[1] = 0; /* trunc trailing spaces */
		if (allspace) outasm[0] = 0;
		if (didcut && outasm[0] == 0) {
			blankline = 1;
		}

		/*           11111111112222222222 */
		/* 012345678901234567890123456789 */
		/* ADDR    xx xx xx xx     INST   */
		if (labelval_pass2 >= 0x00010000) {
			sprintf(tmpbuf, "%04X    ", labelval_pass2 & 0x0000FFFF);
			labelval_pass2 = 0;
		}
		else {
			sprintf(tmpbuf, "%04X    ", locabs_s);
		}
		if (loccnt_s < loccnt_e) sprintf(&tmpbuf[8], "%02X ", (unsigned char)binbuf[loccnt_s]);
		else		       sprintf(&tmpbuf[8], "   ");
		if (loccnt_s + 1 < loccnt_e) sprintf(&tmpbuf[11], "%02X ", (unsigned char)binbuf[loccnt_s + 1]);
		else		       sprintf(&tmpbuf[11], "   ");
		if (loccnt_s + 2 < loccnt_e) sprintf(&tmpbuf[14], "%02X ", (unsigned char)binbuf[loccnt_s + 2]);
		else		       sprintf(&tmpbuf[14], "   ");
		if (loccnt_s + 3 < loccnt_e) sprintf(&tmpbuf[17], "%02X     ", (unsigned char)binbuf[loccnt_s + 3]);
		else		       sprintf(&tmpbuf[17], "        ");
		strcpy(&tmpbuf[24], outasm);
		outasm[0] = 0;
		for (i = 0; i < 256; i++) {
			c = tmpbuf[i];
			if (c == 0) break;
			if (c == 0x0d || c == 0x0a) {
				tmpbuf[i] = 0;
				break;
			}
		}
		if (!tmpbuf[24]) {
			if (blankline) goto skip_pass2;
			blankline = 1;
			tmpbuf[4] = 0;
		}
		else
			blankline = 0;
		if (i >= 256) tmpbuf[255] = 0;
		fprintf(pinfo->pass2, "%s\n", tmpbuf);
		while (loccnt_e - loccnt_s > 4) {
			loccnt_s += 4;
			locabs_s += 4;
			sprintf(tmpbuf, "%04X", locabs_s);
			if (loccnt_s < loccnt_e) sprintf(&tmpbuf[4], "    %02X", (unsigned char)binbuf[loccnt_s]);
			if (loccnt_s + 1 < loccnt_e) sprintf(&tmpbuf[10], " %02X", (unsigned char)binbuf[loccnt_s + 1]);
			if (loccnt_s + 2 < loccnt_e) sprintf(&tmpbuf[13], " %02X", (unsigned char)binbuf[loccnt_s + 2]);
			if (loccnt_s + 3 < loccnt_e) sprintf(&tmpbuf[16], " %02X", (unsigned char)binbuf[loccnt_s + 3]);
			fprintf(pinfo->pass2, "%s\n", tmpbuf);
		}
	skip_pass2:
		if (doflush) {
			goto loop;
		}
	}
	if (errmsg) {

		if (!macro_nest_lev) {
			printf("Error %s%5d: %s\n", get_string(pinfo->filename), pinfo->linenum, errmsg);

#if 0
			printf("%s", linebuf);
			for (i = 0; i < (rp - linebuf); i++) putchar(linebuf[i] == '\t' ? '\t' : ' ');
			printf("^ %s\n", errmsg);
#endif
		}
		else {
			printf("Error %s     : %s\n", get_string(pinfo->filename), errmsg);
		}

		return PRES_ERROR;
	}
	return PRES_OK;
}

int parse_endm(char* linebuf, parseinfo* pi) {
	token t;
	static int nestlev = 0;
	rp = linebuf;
	while (get_token(&t) && t.type != 0) {
		if (t.type == CMD_ENDM) {
			if (nestlev == 0) return 1;
			nestlev--;
			continue;
		}
		else if (t.type == CMD_MACRO) {
			if (!pi->pass2) {
				printf("Error %s%5d: %s\n", get_string(pinfo->filename), pinfo->linenum, EJ("Macros cannot be defined within a macro definition", "}N`Ń}N͒`ł܂"));
			}
			return -1;
		}
		else if (t.type == CMD_REPT) {
			nestlev++;
			if (nestlev >= MAXMACRONEST) {
				if (!pi->pass2) {
					printf("Error %s%5d: %s\n", get_string(pinfo->filename), pinfo->linenum, EJ("Nesting of the REPT is too deep", "REPT̃lXg[܂"));
				}
				return -1;
			}
		}
	}
	return 0;
}

/*---------------------------------------*/
//
// Place unfixed object blocks to abusolute address
//
typedef struct obj_block_t {
	struct obj_block_t* next;
	int		     ns;   // label namespace
	long		     li;   // label index
	int		     size; // block size
} obj_block;
//
// Add a new obj_block into a chain (sorted by size)
//
void add_obj_block(obj_block** head, int ns, long li, int siz) {
	obj_block* onew, * otmp, ** oprev;
	oprev = head;
	otmp = *oprev;
	onew = (obj_block*)malloc(sizeof(obj_block));
	if (onew == NULL) return;
	onew->next = NULL;
	onew->ns = ns;
	onew->li = li;
	onew->size = siz;
	while (otmp) {
		if (otmp->size < siz) {
			*oprev = onew;
			onew->next = otmp;
			return;
		}
		oprev = &(otmp->next);
		otmp = otmp->next;
	}
	*oprev = onew;
}
typedef struct obj_trans_info_t {
	struct obj_trans_info_t* next;
	int  src;  // source location
	int  dest; // destination location
	int  size; // block size
} obj_trans_info;
obj_trans_info* otransinfo;
void add_obj_trans_info(int src, int dest, int siz) {
	obj_trans_info* onew, * otmp, ** oprev;
	oprev = &otransinfo;
	otmp = *oprev;
	onew = (obj_trans_info*)malloc(sizeof(obj_trans_info));
	if (onew == NULL) return;
	onew->next = NULL;
	onew->src = src;
	onew->dest = dest;
	onew->size = siz;
	while (otmp) {
		if (otmp->dest > dest) {
			*oprev = onew;
			onew->next = otmp;
			return;
		}
		oprev = &(otmp->next);
		otmp = otmp->next;
	}
	*oprev = onew;
}

//
// fix abusolute address of object blocks
//
int fix_in_page_address(void) {
	obj_block* page_aligned_blocks; // list of page-aligned blocks, bigger first
	obj_block* in_page_blocks;      // list of in-page blocks, bigger first
	obj_block* o, * on;
	long li;
	int sav_ns, ns;
	label* lp;
	int bin[256];                   // size of each bins (up to 256)
	int binnum;                     // total number of bins
	int locabs_f, locabs_n, la, abs2loc;
	int i, mi, ms, fnd;
	int r;

	sav_ns = get_namespace_id();
	r = 0;
	abs2loc = loccnt - locabs;

	/* process all labels */
	page_aligned_blocks = in_page_blocks = NULL;
	topmost_namespace();
	do {
		ns = get_namespace_id();
		for (li = 1; (lp = getlabelptr(li)) != NULL; li++) {
			if (lp->reftype & LF_PAGE_ALIGNED) {
				add_obj_block(&page_aligned_blocks, ns, li, lp->value2);
				lp->reftype &= ~LF_PAGE_ALIGNED;
			}
			else if (lp->reftype & LF_IN_PAGE) {
				add_obj_block(&in_page_blocks, ns, li, lp->value2);
				lp->reftype &= ~LF_IN_PAGE;
			}
		}
	} while (next_namespace());

	/* initialize first bin */
	locabs_f = locabs_n = locabs & 0xFF00;
	binnum = 0;
	if (locabs & 0x0FF) {
		locabs_n += 0x100;
		bin[binnum++] = 0x100 - (locabs & 0x00FF);
	}
	/* place page_aligned blocks first */
	for (o = page_aligned_blocks; o != NULL; o = o->next) {
		if (locabs_n == 0x10000) {
			goto errxit;
		}
		switch_namespace(o->ns);
		lp = getlabelptr(o->li);
		add_obj_trans_info(lp->value + MAXBINBUF, locabs_n + abs2loc, lp->value2);
		lp->value = locabs_n;
		lp->defstate = LABEL_DEFINED;
		while (o->size > 256) {
			o->size -= 256;
			locabs_n += 0x100;
			bin[binnum++] = 0;
		}
		locabs = locabs_n + o->size;
		loccnt = locabs + abs2loc;
		locabs_n += 0x100;
		bin[binnum++] = 256 - o->size;
	}

	/* put in_page blocks into bin */
	for (o = in_page_blocks; o != NULL; o = o->next) {
		/* search a just-fit bin */
		fnd = 0;
		for (i = 0; i < binnum; i++) {
			if (bin[i] == o->size) {
				switch_namespace(o->ns);
				lp = getlabelptr(o->li);
				la = locabs_f + (i << 8) + (256 - bin[i]);
				add_obj_trans_info(lp->value + MAXBINBUF, la + abs2loc, lp->value2);
				lp->value = la;
				lp->defstate = LABEL_DEFINED;
				bin[i] = 0;
				fnd = 1;
				break;
			}
		}
		if (!fnd) {
			/* search the biggest bin except the last bin */
			ms = 0;
			for (i = 0; i < binnum - 1; i++) {
				if (bin[i] > o->size) {
					if (bin[i] > ms) {
						ms = bin[i];
						mi = i;
					}
				}
			}
			if (ms == 0) {
				/* check the last bin */
				if (bin[binnum - 1] > o->size) {
					mi = binnum - 1;
				}
				else {
					/* no bin can contain it... prepare a new bin */
					if (locabs_n == 0x10000) {
						goto errxit;
					}
					mi = binnum;
					locabs_n += 0x100;
					bin[binnum++] = 256;
				}
			}
			switch_namespace(o->ns);
			lp = getlabelptr(o->li);
			la = locabs_f + (mi << 8) + (256 - bin[mi]);
			add_obj_trans_info(lp->value + MAXBINBUF, la + abs2loc, lp->value2);
			lp->value = la;
			lp->defstate = LABEL_DEFINED;
			bin[mi] -= o->size;
		}
		la = locabs_n - bin[binnum - 1];
		if (la > locabs) {
			locabs = la;
			loccnt = locabs + abs2loc;
		}
	}
	using_in_page_used = 0;
	r = 1;
	/* bin padding info */
	for (i = 0; i < binnum - 1; i++) {
		if (bin[i] == 0) continue;
		printf("Info : %04X-%04X %5d%s\n",
			locabs_f + (i << 8) + 256 - bin[i], locabs_f + (i << 8) + 255, bin[i], EJ(" free space in bytes", "oCg̋󂫗̈"));
	}
errxit:
	if (!r) {
		printf("Error %s\n", EJ("Binary buffer overflow", "oCiobt@܂"));
	}
	/* dispose lists */
	for (o = page_aligned_blocks; o != NULL; o = on) {
		on = o->next;
		free(o);
	}
	for (o = in_page_blocks; o != NULL; o = on) {
		on = o->next;
		free(o);
	}
	switch_namespace(sav_ns);
	return r;
}

//
// place the address-fixed object blocks to the actual location
//
int place_in_page_blocks(void) {
	obj_trans_info* ot, * otn;
	parseinfo pi; // dummy for putbin()
	int i;

	pi.pass2 = NULL; // for putbin()
	pinfo = &pi;
	for (ot = otransinfo; ot != NULL; ot = otn) {
		otn = ot->next;
		while (locabs < ot->dest) putbin(0);
		for (i = 0; i < ot->size; i++) {
			binbuf[ot->dest + i] = binbuf[ot->src + i];
		}
		free(ot);
	}
	otransinfo = NULL;
	return 1;
}

/*---------------------------------------*/

int finish_asm(void) {
	if (!fix_in_page_address()) return 0;
	if (!patch_unres_expr()) return 0;
	if (!place_in_page_blocks()) return 0;
	return 1;
}

int quit_asm(void) {
	patch_unres_expr();
	dispose_unres_expr();
	return 1;
}

int start_asm(void) {
	int i;
	loccnt = locabs = locdol = 0;
	for (i = 0; i < 11; i++) locallabel[i] = 0;
	topmost_namespace();
	db_encode = DB_ASCII_GENERIC;
	for (i = 0; i < 256; i++) mapchr[i] = i;
	rept_num = 0;
	init_using_in_page();
	return 1;
}

unsigned char* getbininfo(int* siz, int* loadaddr, int* execaddr) {
	*siz = loccnt;
	*loadaddr = load_addr;
	*execaddr = exec_addr;
	return binbuf;
}

