/* token.h */

#ifndef _TOKEN_H_
#define _TOKEN_H_

enum {
	/* instructions */
	INST_ADD = 1,
	INST_ADC,
	INST_SUB,
	INST_SBC,
	INST_AND,
	INST_XOR,
	INST_OR,
	INST_CP,
	INST_LD,
	INST_HALT,
	INST_NOP,
	INST_EX,
	INST_DJNZ,
	INST_JR,
	INST_INC,
	INST_DEC,
	INST_RRCA,
	INST_RLCA,
	INST_RRA,
	INST_RLA,
	INST_DAA,
	INST_CPL,
	INST_SCF,
	INST_CCF,
	INST_RET,
	INST_POP,
	INST_EXX,
	INST_JP,
	INST_OUT,
	INST_IN,
	INST_DI,
	INST_EI,
	INST_CALL,
	INST_PUSH,
	INST_RST,
	INST_NEG,
	INST_RETN,
	INST_RETI,
	INST_IM,
	INST_RRD,
	INST_RLD,
	INST_LDI,
	INST_LDD,
	INST_LDIR,
	INST_LDDR,
	INST_CPI,
	INST_CPD,
	INST_CPIR,
	INST_CPDR,
	INST_INI,
	INST_IND,
	INST_INIR,
	INST_INDR,
	INST_OUTI,
	INST_OUTD,
	INST_OTIR,
	INST_OTDR,
	INST_RLC,	// beware the order between RLC and SET
	INST_RRC,
	INST_RL,
	INST_RR,
	INST_SLA,
	INST_SRA,
	INST_SLL,
	INST_SRL,
	INST_BIT,
	INST_RES,
	INST_SET,	// beware the order between RLC and SET
	INST_TSTI,

	/* reserved words */
	REG_A,		// register names
	REG_B,
	REG_C,
	REG_D,
	REG_E,
	REG_H,
	REG_L,
	REG_IXH,
	REG_IXL,
	REG_IYH,
	REG_IYL,
	REG_I,
	REG_R,
	REG_AF,
	REG_BC,
	REG_DE,
	REG_HL,
	REG_SP,
	REG_IX,
	REG_IY,
	REG_AF_,	// AF'
	REG_HL_BRACED,	// (HL)
	REG_DE_BRACED,	// (DE)
	REG_BC_BRACED,	// (BC)
	REG_SP_BRACED,	// (SP)
	REG_IX_BRACED,	// (IX)
	REG_IY_BRACED,	// (IY)
	REG_IX_d_BRACED,	// (IX  ('+' or '-' follows)
	REG_IY_d_BRACED,	// (IY  ('+' or '-' follows)
	REG_C_BRACED,	// (C)
	CND_NZ,		// conditions
	CND_Z,
	CND_NC,
//	CND_C,
#define CND_C REG_C
	CND_PO,
#define CND_NV CND_PO
	CND_PE,
#define CND_V  CND_PE
	CND_P,
	CND_M,

	/* commands for assembler */
	CMD_DB,
	CMD_DW,
	CMD_DS,
	CMD_ORG,
	CMD_MACRO,
	CMD_ENDM,
	CMD_ALIGN,
	CMD_OFFSET,
	CMD_EXTERN,
	CMD_LABELDEF,		// label:	(normal label definition)
	CMD_EQU,		// label EQU x	(normal label definition)
	CMD_R_EQU,		// label = x	(re-definable label)
	CMD_LABELDEF_GLOBAL,	// label::	(global label definition)
	CMD_INCLUDE,
	CMD_BINCLUDE,
	CMD_IF,
	CMD_ELSE,
	CMD_ENDIF,
	CMD_DISPC,
	CMD_ERROR,
	CMD_PROC,
	CMD_ENDP,
	CMD_ELIF,
	CMD_CHRMAP,
	CMD_MAPCHR,
	CMD_ARRANGE,
	CMD_PAGE_ALIGNED,
	CMD_IN_PAGE,
	CMD_REPT,
	CMD_END,

	/* misc */
	T_BRCOPEN,		// '('
	T_BRCCLOSE,		// ')'
	T_COMMA,		// ','
	T_SQUOTE,		// '''
	T_DQUOTE,		// '"'
	T_LOCCNT,		// location counter, '$'

	/* operators */
//	OPR_POS,
	OPR_NEG,
	OPR_NOT,
	OPR_B_NOT,
	OPR_MUL,
	OPR_DIV,
	OPR_MOD,
	OPR_PLUS,
	OPR_MINUS,
	OPR_SHL,
	OPR_SHR,
	OPR_GT,
	OPR_LT,
	OPR_GE,
	OPR_LE,
	OPR_EQ,
	OPR_NE,
	OPR_B_AND,
	OPR_B_XOR,
	OPR_B_OR,
	OPR_AND,
	OPR_OR,
	OPR_DEFINED,

	/* identifier */
	IDENTIFIER,

	/* immidiate */
	IMMIDIATE,
	
	/* deleted text */
	DELTEXT
};
#define INST_START INST_ADD
#define INST_END   INST_TSTI
#define CMD_START  CMD_DB
#define OPR_HL     CMD_LABELDEF

#define MAXTOKENLEN 64

/* type definition */
typedef struct {
	int	type;		/* token type */
	long	value;		/* value of token, or sub info */
	int	tstrlen;	/* length of token string */
	char	*from;		/* where the token is read from */
	char	tstr[MAXTOKENLEN];	/* token string */
} token;

#define IDENTIFIER_REF_ONLY -1

#endif /*_TOKEN_H_*/
